/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Fixed-Point Number Library
 *
 * $Id$
 */

#ifndef FIX_H
#define FIX_H

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "types.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "memalloc.h"

/******************************************************************************\
* Types.
\******************************************************************************/

/* The integral type used to represent a fixed-point number.  This
  type must be capable of representing values from -(2^31) to 2^31-1
  (inclusive). */
typedef int_fast32_t fix_t;

/* The integral type used to respresent higher-precision intermediate results.
  This type should be capable of representing values from -(2^63) to 2^63-1
  (inclusive). */
typedef int_fast64_t fix_big_t;

/******************************************************************************\
* Constants.
\******************************************************************************/

/* The number of bits used for the fractional part of a fixed-point number. */
#define FIX_FRACBITS	14

/* The representation of the value zero. */
#define	FIX_ZERO	0

/* The representation of the value one. */
#define	FIX_ONE		(1 << FIX_FRACBITS)

/* The representation of the value one half. */
#define	FIX_HALF	(1 << (FIX_FRACBITS - 1))

/******************************************************************************\
* Conversion operations.
\******************************************************************************/

/* Convert an int to a fixed-point number. */
#define inttofix(x)	((fix_t)((x) << FIX_FRACBITS))

/* Convert a fixed-point number to an int. */
#define fixtoint(x)	((int)((x) >> FIX_FRACBITS))

/* Convert a fixed-point number to a double. */
#define fixtodbl(x)	(((double)(x)) / (1 << FIX_FRACBITS))

/* Convert a double to a fixed-point number. */
#define dbltofix(x) \
	((fix_t)((x) * (double)(1 << FIX_FRACBITS)))

/******************************************************************************\
* Arithmetic operations.
\******************************************************************************/

/* Calculate the sum of two fixed-point numbers. */
#define	fix_add(x, y)	((x) + (y))

/* Calculate the sum of three fixed-point numbers. */
#define fix_add3(x, y, z)	((x) + (y) + (z))

/* Calculate the difference between two fixed-point numbers. */
#define	fix_sub(x, y)	((x) - (y))

/* Calculate the product of two fixed-point numbers. */
#define	fix_mul(x, y) \
	((fix_t)(((fix_big_t) (x) * (y)) >> FIX_FRACBITS))
#if 0
/* If you don't have a large enough integer type, try this for fix_mul. */
#define	fix_mul(x, y) \
	(((fix_t)(((x) >> 4) * ((y) >> 4))) >> (FIX_FRACBITS - 8))
#endif

/* Calculate the product of a fixed-point number and an int. */
#define	fix_mulbyint(x, y) \
	((x) * (y))

/* Calculate the quotient of two fixed-point numbers. */
#define	fix_div(x, y) \
	((fix_t)(((fix_big_t) (x) << FIX_FRACBITS) / (y)))

/* Negate a fixed-point number. */
#define	fix_neg(x)	(-(x))

/* Add one fixed-point number to another. */
#define	fix_pluseq(x, y) \
	((x) += (y))

/* Subtract one fixed-point number from another. */
#define	fix_minuseq(x, y) \
	((x) -= (y))

/* Multiply one fixed-point number by another. */
#define	fix_muleq(x, y) \
	((x) = fix_mul(x, y))

/******************************************************************************\
* Miscellaneous operations.
\******************************************************************************/

/* Calculate the absolute value of a fixed-point number. */
#define	fix_abs(x)	(((x) >= 0) ? (x) : (-(x)))

/* Is a fixed-point number an integer? */
#define	fix_isint(x) \
	(fix_floor(x) == (x))

/* Get the sign of a fixed-point number. */
#define fix_sgn(x)	((x) >= 0 ? 1 : (-1))

/* Perform an arithmetic shift right of a fixed-point number. */
#define	fix_asr(x, n) \
	((x) >> (n))

/* Perform an arithmetic shift left of a fixed-point number. */
#define	fix_asl(x, n) \
	((x) << (n))

/******************************************************************************\
* Relational operations.
\******************************************************************************/

/* Compare two fixed-point numbers. */
#define fix_cmp(x, y)	((x) > (y) ? 1 : (((x) == (y)) ? 0 : (-1)))

/* Less than. */
#define	fix_lt(x, y)	((x) < (y))

/* Less than or equal. */
#define	fix_lte(x, y)	((x) <= (y))

/* Greater than. */
#define	fix_gt(x, y)	((x) > (y))

/* Greater than or equal. */
#define	fix_gte(x, y)	((x) >= (y))

/******************************************************************************\
* Rounding functions.
\******************************************************************************/

/* Round a fixed-point number to the nearest integer. */
#define	fix_round(x) \
	(((x) < 0) ? fix_floor(fix_add((x), FIX_HALF)) : \
	  fix_neg(fix_floor(fix_add((-(x)), FIX_HALF))))

/* Round a fixed-point number to the nearest integer in the direction of
  negative infinity (i.e., the floor function). */
#define	fix_floor(x) \
	((x) & (~((1 << FIX_FRACBITS) - 1)))

/* Round a fixed-point number to the nearest integer in the direction
  of zero. */
#define fix_trunc(x) \
	(((x) >= 0) ? fix_floor(x) : fix_ceil(x))

#ifdef __cplusplus
}
#endif

#endif
