/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Image Library
 *
 * $Id$
 */

#ifndef IMAGE_H
#define IMAGE_H

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "stream.h"
#include "seq.h"

/******************************************************************************\
* Constants.
\******************************************************************************/

/*
 * Miscellaneous constants.
 */

#define IMAGE_INMEMTHRESH	(64 * 1024 * 1024)
/* The threshold at which image data is no longer stored in memory. */

/*
 * Color models.
 */

#define CM_UNKNOWN	0	/* Unknown color model. */
#define	CM_GRAY		1	/* No color model (i.e., grayscale). */
#define	CM_RGB		2	/* RGB color model. */
#define CM_YCRCB	3	/* YCrCb color model. */

/******************************************************************************\
* Image class and supporting classes.
\******************************************************************************/

/* Image component. */

typedef struct {

	uint_fast32_t tlx_;
	/* The x-coordinate of the top-left corner of the component. */

	uint_fast32_t tly_;
	/* The y-coordinate of the top-left corner of the component. */

	uint_fast32_t hstep_;
	/* The horizontal sampling period in units of the reference grid. */

	uint_fast32_t vstep_;
	/* The vertical sampling period in units of the reference grid. */

	uint_fast32_t width_;
	/* The component width in samples. */

	uint_fast32_t height_;
	/* The component height in samples. */

	uint_fast16_t prec_;
	/* The precision of the sample data (i.e., the number of bits per
	sample).  If the samples are signed values, this quantity
	includes the sign bit. */

	bool sgnd_;
	/* The signedness of the sample data. */

	stream_t *stream_;
	/* The stream containing the component data. */

	int cps_;
	/* The number of characters per sample in the stream. */

} image_cmpt_t;

/* Image. */

typedef struct {

	uint_fast32_t tlx_;
	/* The x-coordinate of the top-left corner of the image bounding box. */

	uint_fast32_t tly_;
	/* The y-coordinate of the top-left corner of the image bounding box. */

	uint_fast32_t brx_;
	/* The x-coordinate of the bottom-right corner of the image bounding
	  box (plus one). */

	uint_fast32_t bry_;
	/* The y-coordinate of the bottom-right corner of the image bounding
	  box (plus one). */

	uint_fast16_t numcmpts_;
	/* The number of components. */

	uint_fast16_t maxcmpts_;
	/* The maximum number of components that this image can have (i.e., the
	  allocated size of the components array). */

	image_cmpt_t **cmpts_;
	/* Per-component information. */

	int colormodel_;
	/* The color model used.  This field is only of particular relevance
	in the case of a multiple component image. */

	bool inmem_;

} image_t;

/* Component parameters. */
/* This data type exists solely for the purposes of the image_create
function. */

typedef struct {

	uint_fast32_t tlx;
	/* The x-coordinate of the top-left corner of the component. */

	uint_fast32_t tly;
	/* The y-coordinate of the top-left corner of the component. */

	uint_fast32_t hstep;
	/* The horizontal sampling period in units of the reference grid. */

	uint_fast32_t vstep;
	/* The vertical sampling period in units of the reference grid. */

	uint_fast32_t width;
	/* The width of the component in samples. */

	uint_fast32_t height;
	/* The height of the component in samples. */

	uint_fast16_t prec;
	/* The precision of the component sample data. */

	bool sgnd;
	/* The signedness of the component sample data. */

} image_cmptparm_t;

/******************************************************************************\
* File format related classes.
\******************************************************************************/

#define	IMAGE_MAXFMTS	32
/* The maximum number of file formats supported. */

/* Image format-dependent operations. */

typedef struct {

	image_t *(*load)(stream_t *in, char *opts);
	/* Load image data from a stream. */

	int (*save)(image_t *image, stream_t *out, char *opts);
	/* Save image data to a stream. */

	int (*validate)(stream_t *in);
	/* Determine if stream data is in a particular format. */

} image_fmtops_t;

/* Image format information. */

typedef struct {

	int id;
	/* The ID for this format. */

	char *name;
	/* The name by which this format is identified. */

	char *ext;
	/* The file name extension associated with this format. */

	image_fmtops_t ops;
	/* The operations for this format. */

} image_fmtinfo_t;

/******************************************************************************\
* Image operations.
\******************************************************************************/

/* Create an image. */
image_t *image_create(uint_fast16_t numcmpts, image_cmptparm_t *cmptparms,
  int colormodel);

image_t *image_create0();

/* Clone an image. */
image_t *image_copy(image_t *image);

/* Deallocate any resources associated with an image. */
void image_destroy(image_t *image);

/* Get the width of the image in units of the image reference grid. */
#define image_width(image) \
	((image)->brx_ - (image)->tlx_)

/* Get the height of the imaage in units of the image reference grid. */
#define	image_height(image) \
	((image)->bry_ - (image)->tly_)

#define image_tlx(image) \
	((image)->tlx_)

#define image_tly(image) \
	((image)->tly_)

#define image_brx(image) \
	((image)->brx_)

#define image_bry(image) \
	((image)->bry_)

/* Get the number of image components. */
#define	image_numcmpts(image) \
	((image)->numcmpts_)

/* Get the color model used by the image. */
#define	image_colormodel(image) \
	((image)->colormodel_)

/* Set the color model for an image. */
void image_setcolormodel(image_t *image, int colormodel);

/* Get the width of a component. */
#define	image_cmptwidth(image, cmptno) \
	((image)->cmpts_[cmptno]->width_)

/* Get the height of a component. */
#define	image_cmptheight(image, cmptno) \
	((image)->cmpts_[cmptno]->height_)

/* Get the format of the sample data. */
#define	image_cmptsgnd(image, cmptno) \
	((image)->cmpts_[cmptno]->sgnd_)

/* Get the precision of the sample data. */
#define	image_cmptprec(image, cmptno) \
	((image)->cmpts_[cmptno]->prec_)

/* Get the horizontal subsampling factor for a component. */
#define	image_cmpthstep(image, cmptno) \
	((image)->cmpts_[cmptno]->hstep_)

/* Get the vertical subsampling factor for a component. */
#define	image_cmptvstep(image, cmptno) \
	((image)->cmpts_[cmptno]->vstep_)

/* Get the x-coordinate of the top-left corner of a component. */
#define	image_cmpttlx(image, cmptno) \
	((image)->cmpts_[cmptno]->tlx_)

/* Get the y-coordinate of the top-left corner of a component. */
#define	image_cmpttly(image, cmptno) \
	((image)->cmpts_[cmptno]->tly_)

/* Get the x-coordinate of the bottom-right corner of a component. */
#define	image_cmptbrx(image, cmptno) \
	((image)->cmpts_[cmptno]->tlx_ + (image)->cmpts_[cmptno]->width_ * (image)->cmpts_[cmptno]->hstep_)

/* Get the y-coordinate of the bottom-right corner of a component. */
#define	image_cmptbry(image, cmptno) \
	((image)->cmpts_[cmptno]->tly_ + (image)->cmpts_[cmptno]->height_ * (image)->cmpts_[cmptno]->vstep_)

/* Get the raw size of an image (i.e., the nominal size of the image without
  any compression. */
uint_fast32_t image_getrawsize(image_t *image);

/* Create an image from a stream in some specified format. */
image_t *image_load(stream_t *in, int fmt, char *optstr);

/* Write an image to a stream in a specified format. */
int image_save(image_t *image, stream_t *out, int fmt, char *optstr);

/* Read a rectangular region of an image component. */
/* The position and size of the rectangular region to be read is specified
relative to the component's coordinate system. */
int image_readcmpt(image_t *image, uint_fast16_t cmptno, uint_fast32_t x,
  uint_fast32_t y, uint_fast32_t width, uint_fast32_t height, matrix_t *data);

/* Write a rectangular region of an image component. */
int image_writecmpt(image_t *image, uint_fast16_t cmptno, uint_fast32_t x,
  uint_fast32_t y, uint_fast32_t width, uint_fast32_t height, matrix_t *data);

/* Delete a component from an image. */
void image_delcmpt(image_t *image, uint_fast16_t cmptno);

/* Add a component to an image. */
int image_addcmpt(image_t *image, uint_fast16_t cmptno,
  image_cmptparm_t *cmptparm);

/* Copy a component from one image to another. */
int image_copycmpt(image_t *dstimage, int dstcmptno, image_t *srcimage,
  int srccmptno);

/******************************************************************************\
* Image format-related operations.
\******************************************************************************/

/* Clear the table of image formats. */
void image_clearfmts();

/* Add entry to table of image formats. */
int image_addfmt(int id, char *name, char *ext, image_fmtops_t *ops);

/* Get the ID for the image format with the specified name. */
int image_strtofmt(char *s);

/* Get the name of the image format with the specified ID. */
char *image_fmttostr(int fmt);

/* Guess the format of an image file based on its name. */
int image_fmtfromname(char *filename);

/* Get the format of image data in a stream. */
int image_getfmt(stream_t *in);

/******************************************************************************\
* Image format-dependent operations.
\******************************************************************************/

/* Format-dependent operations for MIF support. */
image_t *image_load_mif(stream_t *in, char *optstr);
int image_save_mif(image_t *image, stream_t *out, char *optstr);
int mif_validate(stream_t *in);

/* Format-dependent operations for PNM support. */
image_t *image_load_pnm(stream_t *in, char *optstr);
int image_save_pnm(image_t *image, stream_t *out, char *optstr);
int pnm_validate(stream_t *in);

/* Format-dependent operations for Sun Rasterfile support. */
image_t *image_load_ras(stream_t *in, char *optstr);
int image_save_ras(image_t *image, stream_t *out, char *optstr);
int ras_validate(stream_t *in);

/* Format-dependent operations for BMP support. */
image_t *image_load_bmp(stream_t *in, char *optstr);
int image_save_bmp(image_t *image, stream_t *out, char *optstr);
int bmp_validate(stream_t *in);

/* Format-dependent operations for JP2 support. */
image_t *image_load_jp2(stream_t *in, char *optstr);
int image_save_jp2(image_t *image, stream_t *out, char *optstr);
int jp2_validate(stream_t *in);

/* Format-dependent operations for JPEG-2000 code stream support. */
image_t *image_load_jpc(stream_t *in, char *optstr);
int image_save_jpc(image_t *image, stream_t *out, char *optstr);
int jpc_validate(stream_t *in);

#ifdef __cplusplus
}
#endif

#endif
