/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Sequence/Matrix Library
 *
 * $Id$
 */

#ifndef SEQ_H
#define SEQ_H

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "stream.h"
#include "fix.h"

/******************************************************************************\
* Constants.
\******************************************************************************/

#define MATRIX_REF	0x0001

/******************************************************************************\
* Types.
\******************************************************************************/

/* Matrix. */

typedef struct {

	int flags_;
	/* Additional state information. */

	int_fast32_t xstart_;
	/* The starting horizontal index. */

	int_fast32_t ystart_;
	/* The starting vertical index. */

	int_fast32_t xend_;
	/* The ending horizontal index. */

	int_fast32_t yend_;
	/* The ending vertical index. */

	int_fast32_t numrows_;
	/* The number of rows in the matrix. */

	int_fast32_t numcols_;
	/* The number of columns in the matrix. */

	fix_t **rows_;
	/* Pointers to the start of each row. */

	int_fast32_t maxrows_;
	/* The allocated size of the rows array. */

	fix_t *data_;
	/* The matrix data buffer. */

	int_fast32_t datasize_;
	/* The allocated size of the data array. */

} matrix_t;

typedef matrix_t seq2d_t;
typedef matrix_t seq_t;

/******************************************************************************\
* Functions.
\******************************************************************************/

/* Get the number of rows. */
#define matrix_numrows(matrix) \
	((matrix)->numrows_)

/* Get the number of columns. */
#define matrix_numcols(matrix) \
	((matrix)->numcols_)

/* Get a matrix element. */
#define matrix_get(matrix, i, j) \
	((matrix)->rows_[i][j])

/* Set a matrix element. */
#define matrix_set(matrix, i, j, v) \
	((matrix)->rows_[i][j] = (v))

/* Get an element from a matrix that is known to be a row or column vector. */
#define matrix_getv(matrix, i) \
	(((matrix)->numrows_ == 1) ? ((matrix)->rows_[0][i]) : \
	  ((matrix)->rows_[i][0]))

/* Set an element in a matrix that is known to be a row or column vector. */
#define matrix_setv(matrix, i, v) \
	(((matrix)->numrows_ == 1) ? ((matrix)->rows_[0][i] = (v)) : \
	  ((matrix)->rows_[i][0] = (v)))

/* Get the address of an element in a matrix. */
#define	matrix_getref(matrix, i, j) \
	(&(matrix)->rows_[i][j])

#define	matrix_getvref(matrix, i) \
	(((matrix)->numrows_ > 1) ? matrix_getref(matrix, i, 0) : matrix_getref(matrix, 0, i))

#define matrix_length(matrix) \
	(max((matrix)->numrows_, (matrix)->numcols_))

/* Create a matrix with the specified dimensions. */
matrix_t *matrix_create(int numrows, int numcols);

/* Destroy a matrix. */
void matrix_destroy(matrix_t *matrix);

/* Resize a matrix.  The previous contents of the matrix are lost. */
int matrix_resize(matrix_t *matrix, int numrows, int numcols);

int matrix_outputraw(matrix_t *matrix, stream_t *out);

int matrix_output(matrix_t *matrix, FILE *out);

matrix_t *matrix_inputraw(stream_t *in);

/* Create a matrix that references part of another matrix. */
void matrix_bindsub(matrix_t *mat0, matrix_t *mat1, int r0, int c0,
  int r1, int c1);

/* Create a matrix that is a reference to a row of another matrix. */
#define matrix_bindrow(mat0, mat1, r) \
  (matrix_bindsub((mat0), (mat1), (r), 0, (r), (mat1)->numcols_ - 1))

/* Create a matrix that is a reference to a column of another matrix. */
#define matrix_bindcol(mat0, mat1, c) \
  (matrix_bindsub((mat0), (mat1), 0, (c), (mat1)->numrows_ - 1, (c)))

void matrix_clip(matrix_t *matrix, fix_t minval, fix_t maxval);

void matrix_asl(matrix_t *matrix, int n);
void matrix_asr(matrix_t *matrix, int n);
void matrix_divpow2(matrix_t *matrix, int n);

void matrix_setall(matrix_t *matrix, fix_t val);

void matrix_andeq(matrix_t *matrix, fix_t val);

#define	matrix_rowstep(matrix) \
	(((matrix)->numrows_ > 1) ? ((matrix)->rows_[1] - (matrix)->rows_[0]) : (0))

#define	matrix_step(matrix) \
	(((matrix)->numrows_ > 1) ? (matrix_rowstep(matrix)) : (1))

#define	seq2d_xstart(s) \
	((s)->xstart_)
#define	seq2d_ystart(s) \
	((s)->ystart_)
#define	seq2d_xend(s) \
	((s)->xend_)
#define	seq2d_yend(s) \
	((s)->yend_)
#define	seq2d_getref(s, x, y) \
	(matrix_getref(s, (y) - (s)->ystart_, (x) - (s)->xstart_))
#define	seq2d_get(s, x, y) \
	(matrix_get(s, (y) - (s)->ystart_, (x) - (s)->xstart_))
#define	seq2d_rowstep(s) \
	matrix_rowstep(s)
#define	seq2d_width(s) \
	((s)->xend_ - (s)->xstart_)
#define	seq2d_height(s) \
	((s)->yend_ - (s)->ystart_)

matrix_t *seq2d_create(int xstart, int ystart, int xend, int yend);

#define	seq2d_setshift(s, x, y) \
	((s)->xstart_ = (x), (s)->ystart_ = (y), \
	  (s)->xend_ = (s)->xstart_ + (s)->numcols_, \
	  (s)->yend_ = (s)->ystart_ + (s)->numrows_)

#define	seq_create(start, end) \
	(seq2d_create(start, 0, end, 1))

#define	seq_destroy(seq) \
	(seq2d_destroy(seq))

void seq2d_bindsub(matrix_t *s, matrix_t *s1, int xstart, int ystart,
  int xend, int yend);

#define	seq2d_destroy(s) \
	matrix_destroy(s)

#define seq_set(seq, i, v) \
	((seq)->rows_[0][(i) - (seq)->xstart_] = (v))
#define	seq_getref(seq, i) \
	(&(seq)->rows_[0][(i) - (seq)->xstart_])
#define	seq_get(seq, i) \
	((seq)->rows_[0][(i) - (seq)->xstart_])
#define	seq_start(seq) \
	((seq)->xstart_)
#define	seq_end(seq) \
	((seq)->xend_)

seq_t *seq_upsample(seq_t *seq, int n);
seq_t *seq_conv(seq_t *seq0, seq_t *seq1);

seq2d_t *seq2d_copy(seq2d_t *x);
int matrix_cmp(matrix_t *mat0, matrix_t *mat1);
fix_t seq_norm(seq_t *x);

#ifdef __cplusplus
}
#endif

#endif
