/*
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * JP2 Library
 *
 * $Id$
 */

#include <stream.h>
#include <stdlib.h>
#include "memalloc.h"
#include "jp2.h"

static int jp2_getuint8(stream_t *in, uint_fast8_t *val);
static int jp2_getuint16(stream_t *in, uint_fast16_t *val);
static int jp2_getuint32(stream_t *in, uint_fast32_t *val);
static int jp2_getuint64(stream_t *in, uint_fast64_t *val);
static int jp2_putuint8(stream_t *out, uint_fast8_t val);
static int jp2_putuint16(stream_t *out, uint_fast16_t val);
static int jp2_putuint32(stream_t *out, uint_fast32_t val);
static int jp2_putuint64(stream_t *out, uint_fast64_t val);

jp2_box_t *jp2_box_get(stream_t *in);

static int jp2_jp_getdata(jp2_box_t *box, stream_t *in);
static int jp2_jp_putdata(jp2_box_t *box, stream_t *out);
static int jp2_ftyp_getdata(jp2_box_t *box, stream_t *in);
static int jp2_ftyp_putdata(jp2_box_t *box, stream_t *out);
static int jp2_ihdr_getdata(jp2_box_t *box, stream_t *in);
static int jp2_ihdr_putdata(jp2_box_t *box, stream_t *out);
static void jp2_bpcc_destroy(jp2_box_t *box);
static int jp2_bpcc_getdata(jp2_box_t *box, stream_t *in);
static int jp2_bpcc_putdata(jp2_box_t *box, stream_t *out);
static int jp2_colr_getdata(jp2_box_t *box, stream_t *in);
static int jp2_colr_putdata(jp2_box_t *box, stream_t *out);

jp2_boxinfo_t jp2_boxinfos[] = {
	{JP2_BOX_JP, "Signature (JP)", 0,
	  {0, 0, jp2_jp_getdata, jp2_jp_putdata}},
	{JP2_BOX_FTYP, "File Type (FTYP)", 0,
	  {0, 0, jp2_ftyp_getdata, jp2_ftyp_putdata}},
	{JP2_BOX_JP2H, "JP2 Header (JP2H)", JP2_BOX_SUPER,
	  {0, 0, 0, 0}},
	{JP2_BOX_IHDR, "Image Header (IHDR)", 0,
	  {0, 0, jp2_ihdr_getdata, jp2_ihdr_putdata}},
	{JP2_BOX_BPCC, "Bits Per Component (BPCC)", 0,
	  {0, jp2_bpcc_destroy, jp2_bpcc_getdata, jp2_bpcc_putdata}},
	{JP2_BOX_COLR, "Color Specification (COLR)", 0,
	  {0, 0, jp2_colr_getdata, jp2_colr_putdata}},
	{JP2_BOX_PCLR, "Palette (PCLR)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_CMAP, "Component Mapping (CMAP)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_CDEF, "Channel Definition (CDEF)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_RES, "Resolution (RES)", JP2_BOX_SUPER,
	  {0, 0, 0, 0}},
	{JP2_BOX_RESC, "Capture Resolution (RESC)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_RESD, "Default Display Resolution (RESD)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_JP2C, "Contiguous Code Stream (JP2C)", JP2_BOX_NODATA,
	  {0, 0, 0, 0}},
	{JP2_BOX_JP2I, "Intellectual Property (JP2I)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_XML, "XML (XML)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_UUID, "UUID (UUID)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_UINF, "UUID Info (UINF)", JP2_BOX_SUPER,
	  {0, 0, 0, 0}},
	{JP2_BOX_ULST, "UUID List (ULST)", 0,
	  {0, 0, 0, 0}},
	{JP2_BOX_URL, "URL (URL)", 0,
	  {0, 0, 0, 0}},
	{0, 0, 0, {0, 0, 0, 0}},
};

jp2_boxinfo_t jp2_boxinfo_unk = {
	0, "Unknown", 0, {0, 0, 0, 0}
};

jp2_boxinfo_t *jp2_boxinfolookup(int type)
{
	jp2_boxinfo_t *boxinfo;
	for (boxinfo = jp2_boxinfos; boxinfo->name; ++boxinfo) {
		if (boxinfo->type == type) {
			return boxinfo;
		}
	}
	return &jp2_boxinfo_unk;
}

jp2_box_t *jp2_box_create(int type)
{
	jp2_box_t *box;
	jp2_boxinfo_t *boxinfo;

	if (!(box = malloc(sizeof(jp2_box_t)))) {
		return 0;
	}
	memset(box, 0, sizeof(jp2_box_t));
	box->type = type;
	box->len = 0;
	if (!(boxinfo = jp2_boxinfolookup(type))) {
		return 0;
	}
	box->info = boxinfo;
	box->ops = &boxinfo->ops;
	return box;
}

/******************************************************************************\
* Box destructor.
\******************************************************************************/

void jp2_box_destroy(jp2_box_t *box)
{
	if (box->ops->destroy) {
		(*box->ops->destroy)(box);
	}
	free(box);
}

void jp2_bpcc_destroy(jp2_box_t *box)
{
	jp2_bpcc_t *bpcc = &box->data.bpcc;
	if (bpcc->bpcs) {
		free(bpcc->bpcs);
		bpcc->bpcs = 0;
	}
}

/******************************************************************************\
* Box input.
\******************************************************************************/

jp2_box_t *jp2_box_get(stream_t *in)
{
	jp2_box_t *box;
	jp2_boxinfo_t *boxinfo;
	stream_t *tmpstream;
	uint_fast32_t len;
	uint_fast64_t extlen;
	bool dataflag;

	box = 0;
	tmpstream = 0;

	if (!(box = malloc(sizeof(jp2_box_t)))) {
		goto error;
	}
	box->ops = &jp2_boxinfo_unk.ops;
	if (jp2_getuint32(in, &len) || jp2_getuint32(in, &box->type)) {
		goto error;
	}
	boxinfo = jp2_boxinfolookup(box->type);
	box->info = boxinfo;
	box->ops = &boxinfo->ops;
	box->len = len;
	if (box->len == 1) {
		if (jp2_getuint64(in, &extlen)) {
			goto error;
		}
		box->len = extlen;
	}
	if (box->len != 0 && box->len < 8) {
		goto error;
	}

	dataflag = !(box->info->flags & (JP2_BOX_SUPER | JP2_BOX_NODATA));

	if (dataflag) {
		if (!(tmpstream = stream_memopen(0, 0))) {
			goto error;
		}
		if (stream_copy(tmpstream, in, box->len - JP2_BOX_HDRLEN)) {
			goto error;
		}
		stream_rewind(tmpstream);

		if (box->ops->getdata) {
			if ((*box->ops->getdata)(box, tmpstream)) {
				goto error;
			}
		}
		stream_close(tmpstream);
	}

	return box;
	abort();

error:
	if (box) {
		jp2_box_destroy(box);
	}
	if (tmpstream) {
		stream_close(tmpstream);
	}
	return 0;
}

static int jp2_jp_getdata(jp2_box_t *box, stream_t *in)
{
	jp2_jp_t *jp = &box->data.jp;
	if (jp2_getuint32(in, &jp->magic)) {
		return -1;
	}
	return 0;
}

static int jp2_ftyp_getdata(jp2_box_t *box, stream_t *in)
{
	jp2_ftyp_t *ftyp = &box->data.ftyp;
	int i;
	if (jp2_getuint32(in, &ftyp->majver) || jp2_getuint32(in, &ftyp->minver)) {
		return -1;
	}
	ftyp->numcompatcodes = ((box->len - JP2_BOX_HDRLEN) - 8) / 4;
	if (ftyp->numcompatcodes > JP2_FTYP_MAXCOMPATCODES) {
		return -1;
	}
	for (i = 0; i < ftyp->numcompatcodes; ++i) {
		if (jp2_getuint32(in, &ftyp->compatcodes[i])) {
			return -1;
		}
	}
	return 0;
}

static int jp2_ihdr_getdata(jp2_box_t *box, stream_t *in)
{
	jp2_ihdr_t *ihdr = &box->data.ihdr;
	if (jp2_getuint32(in, &ihdr->height) || jp2_getuint32(in, &ihdr->width) ||
	  jp2_getuint16(in, &ihdr->numcmpts) || jp2_getuint8(in, &ihdr->bpc) ||
	  jp2_getuint8(in, &ihdr->comptype) || jp2_getuint8(in, &ihdr->csunk) ||
	  jp2_getuint8(in, &ihdr->ipr)) {
		return -1;
	}
	return 0;
}

static int jp2_bpcc_getdata(jp2_box_t *box, stream_t *in)
{
	jp2_bpcc_t *bpcc = &box->data.bpcc;
	int i;
	bpcc->numcmpts = box->len - JP2_BOX_HDRLEN;
	if (!(bpcc->bpcs = malloc(bpcc->numcmpts * sizeof(uint_fast8_t)))) {
		return -1;
	}
	for (i = 0; i < bpcc->numcmpts; ++i) {
		if (jp2_getuint8(in, &bpcc->bpcs[i])) {
			return -1;
		}
	}
	return 0;
}

static int jp2_colr_getdata(jp2_box_t *box, stream_t *in)
{
	jp2_colr_t *colr = &box->data.colr;
	if (jp2_getuint8(in, &colr->method) || jp2_getuint8(in, &colr->pri) ||
	  jp2_getuint8(in, &colr->approx)) {
		return -1;
	}
	switch (colr->method) {
	case JP2_COLR_ENUM:
		if (jp2_getuint32(in, &colr->csid)) {
			return -1;
		}
		break;
	case JP2_COLR_ICC:
		/* XXX - not implemented */
		abort();
		break;
	}
	return 0;
}

/******************************************************************************\
* Box output.
\******************************************************************************/

int jp2_box_put(jp2_box_t *box, stream_t *out)
{
	stream_t *tmpstream;
	bool extlen;
	bool dataflag;

	tmpstream = 0;

	dataflag = !(box->info->flags & (JP2_BOX_SUPER | JP2_BOX_NODATA));

	if (dataflag) {
		tmpstream = stream_memopen(0, 0);
		if (box->ops->putdata) {
			if ((*box->ops->putdata)(box, tmpstream)) {
				goto error;
			}
		}
		box->len = stream_tell(tmpstream) + JP2_BOX_HDRLEN;
		stream_rewind(tmpstream);
	}
	extlen = (box->len >= (((uint_fast64_t)1) << 32)) != 0;
	if (jp2_putuint32(out, extlen ? 1 : box->len)) {
		goto error;
	}
	if (jp2_putuint32(out, box->type)) {
		goto error;
	}
	if (extlen) {
		if (jp2_putuint64(out, box->len)) {
			goto error;
		}
	}

	if (dataflag) {
		if (stream_copy(out, tmpstream, box->len - JP2_BOX_HDRLEN)) {
			goto error;
		}
		stream_close(tmpstream);
	}

	return 0;
	abort();

error:

	if (tmpstream) {
		stream_close(tmpstream);
	}
	return -1;
}

static int jp2_jp_putdata(jp2_box_t *box, stream_t *out)
{
	jp2_jp_t *jp = &box->data.jp;
	if (jp2_putuint32(out, jp->magic)) {
		return -1;
	}
	return 0;
}

static int jp2_ftyp_putdata(jp2_box_t *box, stream_t *out)
{
	jp2_ftyp_t *ftyp = &box->data.ftyp;
	int i;
	if (jp2_putuint32(out, ftyp->majver) || jp2_putuint32(out, ftyp->minver)) {
		return -1;
	}
	for (i = 0; i < ftyp->numcompatcodes; ++i) {
		if (jp2_putuint32(out, ftyp->compatcodes[i])) {
			return -1;
		}
	}
	return 0;
}

static int jp2_ihdr_putdata(jp2_box_t *box, stream_t *out)
{
	jp2_ihdr_t *ihdr = &box->data.ihdr;
	if (jp2_putuint32(out, ihdr->height) || jp2_putuint32(out, ihdr->width) ||
	  jp2_putuint16(out, ihdr->numcmpts) || jp2_putuint8(out, ihdr->bpc) ||
	  jp2_putuint8(out, ihdr->comptype) || jp2_putuint8(out, ihdr->csunk) ||
	  jp2_putuint8(out, ihdr->ipr)) {
		return -1;
	}
	return 0;
}

static int jp2_bpcc_putdata(jp2_box_t *box, stream_t *out)
{
	jp2_bpcc_t *bpcc = &box->data.bpcc;
	int i;
	for (i = 0; i < bpcc->numcmpts; ++i) {
		if (jp2_putuint8(out, bpcc->bpcs[i])) {
			return -1;
		}
	}
	return 0;
}

static int jp2_colr_putdata(jp2_box_t *box, stream_t *out)
{
	jp2_colr_t *colr = &box->data.colr;
	if (jp2_putuint8(out, colr->method) || jp2_putuint8(out, colr->pri) ||
	  jp2_putuint8(out, colr->approx)) {
		return -1;
	}
	switch (colr->method) {
	case JP2_COLR_ENUM:
		if (jp2_putuint32(out, colr->csid)) {
			return -1;
		}
		break;
	case JP2_COLR_ICC:
		/* XXX - not implemented */
		abort();
		break;
	}
	return 0;
}

/******************************************************************************\
* Input operations for primitive types.
\******************************************************************************/

static int jp2_getuint8(stream_t *in, uint_fast8_t *val)
{
	int c;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	if (val) {
		*val = c;
	}
	return 0;
}

static int jp2_getuint16(stream_t *in, uint_fast16_t *val)
{
	uint_fast16_t v;
	int c;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = c;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = (v << 8) | c;
	if (val) {
		*val = v;
	}
	return 0;
}

static int jp2_getuint32(stream_t *in, uint_fast32_t *val)
{
	uint_fast32_t v;
	int c;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = c;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = (v << 8) | c;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = (v << 8) | c;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = (v << 8) | c;
	if (val) {
		*val = v;
	}
	return 0;
}

static int jp2_getuint64(stream_t *in, uint_fast64_t *val)
{
	abort();
}

/******************************************************************************\
* Output operations for primitive types.
\******************************************************************************/

static int jp2_putuint8(stream_t *out, uint_fast8_t val)
{
	if (stream_putc(out, val & 0xff) == EOF) {
		return -1;
	}
	return 0;
}

static int jp2_putuint16(stream_t *out, uint_fast16_t val)
{
	if (stream_putc(out, (val >> 8) & 0xff) == EOF ||
	  stream_putc(out, val & 0xff) == EOF) {
		return -1;
	}
	return 0;
}

static int jp2_putuint32(stream_t *out, uint_fast32_t val)
{
	if (stream_putc(out, (val >> 24) & 0xff) == EOF ||
	  stream_putc(out, (val >> 16) & 0xff) == EOF ||
	  stream_putc(out, (val >> 8) & 0xff) == EOF ||
	  stream_putc(out, val & 0xff) == EOF) {
		return -1;
	}
	return 0;
}

static int jp2_putuint64(stream_t *out, uint_fast64_t val)
{
	if (jp2_putuint32(out, (val >> 32) & 0xffffffffUL) ||
	  jp2_putuint32(out, val & 0xffffffffUL)) {
		return -1;
	}
	return 0;
}
