/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * JPEG-2000 Code Stream Library
 *
 * $Id$
 */

#ifndef CS_H
#define CS_H

#include "jpccod.h"
#include "image.h"
#include "stream.h"

/******************************************************************************\
* Constants and Types.
\******************************************************************************/

#define JPC_MAXLYRS	16384

/**************************************\
* Code stream.
\**************************************/

#define	JPC_CS_INIT	0
#define	JPC_CS_MHDR	1
#define	JPC_CS_THDR	2
#define	JPC_CS_MTLR	3
#define	JPC_CS_TDATA	4

#define	JPC_CS_RD	0x01
#define	JPC_CS_WR	0x02
#define	JPC_CS_EIO	0x04
#define	JPC_CS_ELIMIT	0x08
#define	JPC_CS_EOF	0x10

/*
 * Sadly, the code stream syntax was not designed in such a way that any
 * given marker segment can be correctly decoded without additional state
 * derived from previously decoded marker segments.
 *
 * A RGN/COC/QCC marker segment cannot be decoded unless the number of
 * components is known.
 */

typedef struct {

	int numcomps;
	/* The number of components. */

} jpc_cstate_t;

/**************************************\
* SOT marker segment parameters.
\**************************************/

typedef struct {

	uint_fast16_t tileno;
	/* The tile number. */

	uint_fast32_t len;
	/* The combined length of the marker segment and its auxilary data
	  (i.e., packet data). */

	uint_fast8_t partno;
	/* The tile-part instance. */

	uint_fast8_t numparts;
	/* The number of tile-parts. */

} jpc_sot_t;

/**************************************\
* SIZ marker segment parameters.
\**************************************/

typedef struct {

	uint_fast8_t prec;
	/* The precision of the samples. */

	uint_fast8_t sgnd;
	/* The signedness of the samples. */

	uint_fast8_t hsamp;
	/* The horizontal separation of samples with respect to the reference
	  grid. */

	uint_fast8_t vsamp;
	/* The vertical separation of samples with respect to the reference
	  grid. */

} jpc_sizcomp_t;

typedef struct {

	uint_fast16_t caps;
	/* The code stream capabilities. */

	uint_fast32_t width;
	/* The width of the image in units of the reference grid. */

	uint_fast32_t height;
	/* The height of the image in units of the reference grid. */

	uint_fast32_t xoff;
	/* The horizontal offset from the origin of the reference grid to the
	  left side of the image area. */

	uint_fast32_t yoff;
	/* The vertical offset from the origin of the reference grid to the
	  top side of the image area. */

	uint_fast32_t tilewidth;
	/* The nominal width of a tile in units of the reference grid. */

	uint_fast32_t tileheight;
	/* The nominal height of a tile in units of the reference grid. */

	uint_fast32_t tilexoff;
	/* The horizontal offset from the origin of the reference grid to the
	  left side of the first tile. */

	uint_fast32_t tileyoff;
	/* The vertical offset from the origin of the reference grid to the
	  top side of the first tile. */

	uint_fast16_t numcomps;
	/* The number of components. */

	jpc_sizcomp_t *comps;
	/* The per-component information. */

} jpc_siz_t;

/**************************************\
* COD marker segment parameters.
\**************************************/

/* Coding style constants. */
#define	JPC_COX_PRT	0x01 /* Precincts may be used. */
#define	JPC_COD_SOP	0x02 /* SOP marker segments may be used. */
#define	JPC_COD_EPH	0x04 /* EPH marker segments may be used. */

/* Progression order constants. */
#define	JPC_COD_LRCPPRG	0
	  /* Layer-resolution-component-precinct progressive (i.e., progressive by fidelity). */
#define	JPC_COD_RLCPPRG	1
	  /* Resolution-layer-component-precinct progressive (i.e., progressive by resolution). */
#define	JPC_COD_RPCLPRG	2
	  /* Resolution-precinct-component-layer progressive. */
#define	JPC_COD_PCRLPRG	3
	  /* Precinct-component-resolution-layer progressive. */
#define	JPC_COD_CPRLPRG	4
	  /* Component-position-resolution-layer progressive. */

/* Code block style constants. */
#define	JPC_COX_LAZY	0x01 /* Selective arithmetic coding bypass. */
#define	JPC_COX_RESET	0x02 /* Reset context probabilities. */
#define	JPC_COX_TERMALL	0x04 /* Terminate all coding passes. */
#define	JPC_COX_VSC	0x08 /* Vertical stripe causal context formation. */
#define	JPC_COX_PTERM	0x10 /* Predictable termination. */
#define	JPC_COX_SEGSYM	0x20 /* Use segmentation symbols. */

/* Transform constants. */
#define	JPC_COX_INS	0x00 /* Irreversible 9/7. */
#define	JPC_COX_RFT	0x01 /* Reversible 5/3. */

/* Multicomponent transform constants. */
#define	JPC_COD_NOMCT	0x00 /* No multicomponent transform. */
#define	JPC_COD_MCT	0x01 /* Multicomponent transform. */

#define	JPC_COX_CBLKSIZEEXPN(x)		((x) - 2)
#define	JPC_COX_GETCBLKSIZEEXPN(x)	((x) + 2)

typedef struct {

	uint_fast8_t parwidthval;
	/* The packet partition width. */

	uint_fast8_t parheightval;
	/* The packet partition height. */

} jpc_coxrlvl_t;

typedef struct {

	uint_fast8_t csty;
	/* The coding style. */

	uint_fast8_t numdlvls;
	/* The number of decomposition levels. */

	uint_fast8_t cblkwidthval;
	/* The nominal code block width specifier. */

	uint_fast8_t cblkheightval;
	/* The nominal code block height specifier. */

	uint_fast8_t cblksty;
	/* The style of coding passes. */

	uint_fast8_t qmfbid;
	/* The QMFB employed. */

	int numrlvls;
	/* The number of resolution levels. */

	jpc_coxrlvl_t rlvls[JPC_MAXRLVLS];
	/* The per-resolution-level information. */

} jpc_coxcp_t;

typedef struct {

	uint_fast8_t csty;

	uint_fast8_t prg;
	/* The progression order. */

	uint_fast16_t numlyrs;
	/* The number of layers. */

	uint_fast8_t mctrans;
	/* The multicomponent transform. */

	jpc_coxcp_t compparms;
	/* Component-related parameters. */

} jpc_cod_t;

typedef struct {

	uint_fast16_t compno;
	/* The component number. */
	/* This member is only used for COC marker segments. */

	jpc_coxcp_t compparms;
	/* Component-related parameters. */

} jpc_coc_t;

/**************************************\
* RGN marker segment parameters.
\**************************************/

#define	JPC_RGN_MAXSHIFT	0x00

typedef struct {

	uint_fast16_t compno;
	/* The component to which the marker applies. */

	uint_fast8_t roisty;
	/* The ROI style. */

	uint_fast8_t roishift;
	/* The ROI shift value. */

} jpc_rgn_t;

/**************************************\
* QCD/QCC marker segment parameters.
\**************************************/

/* Quantization style constants. */
#define	JPC_QCX_NOQNT	0 /* No quantization. */
#define	JPC_QCX_SIQNT	1 /* Scalar quantization, implicit. */
#define	JPC_QCX_SEQNT	2 /* Scalar quantization, explicit. */

/* Stepsize manipulation macros. */
#define	JPC_QCX_GETEXPN(x)	((x) >> 11)
#define	JPC_QCX_GETMANT(x)	((x) & 0x07ff)
#define	JPC_QCX_EXPN(x)		(assert(!((x) & (~0x1f))), (((x) & 0x1f) << 11))
#define	JPC_QCX_MANT(x)		(assert(!((x) & (~0x7ff))), ((x) & 0x7ff))

typedef struct {

	uint_fast8_t qntsty;
	/* The quantization style. */

	int numstepsizes;
	/* The number of step sizes. */

	uint_fast16_t *stepsizes;
	/* The step sizes. */

	uint_fast8_t numguard;
	/* The number of guard bits. */

} jpc_qcxcp_t;

typedef struct {

	uint_fast16_t compno;
	/* The component associated with this marker segment. */

	jpc_qcxcp_t compparms;

} jpc_qcc_t;

typedef struct {

	jpc_qcxcp_t compparms;

} jpc_qcd_t;

/**************************************\
* POD marker segment parameters.
\**************************************/

typedef struct {

	uint_fast8_t prgord;
	/* The progression order. */

	uint_fast8_t rlvlnostart;
	/* The lower bound (inclusive) on the resolution level for the
	  progression order volume. */

	uint_fast8_t rlvlnoend;
	/* The upper bound (exclusive) on the resolution level for the
	  progression order volume. */

	uint_fast16_t compnostart;
	/* The lower bound (inclusive) on the component for the progression
	  order volume. */

	uint_fast16_t compnoend;
	/* The upper bound (exclusive) on the component for the progression
	  order volume. */

	uint_fast16_t lyrnoend;
	/* The upper bound (exclusive) on the layer for the progression
	  order volume. */

} jpc_pocpchg_t;

typedef jpc_pocpchg_t pchg_t;

typedef struct {

	int numpchgs;
	/* The number of progression order changes. */

	jpc_pocpchg_t *pchgs;
	/* The per-progression-order-change information. */

} jpc_poc_t;

/**************************************\
* PPM/PPT marker segment parameters.
\**************************************/

typedef struct {

	uint_fast8_t ind;

	uint_fast16_t len;

	unsigned char *data;

} jpc_ppm_t;

typedef struct {

	uint_fast8_t ind;

	uint_fast32_t len;

	unsigned char *data;

} jpc_ppt_t;

/**************************************\
* COM marker segment parameters.
\**************************************/

/* Registration ID constants. */
#define	JPC_COM_BIN	0x00
#define	JPC_COM_LATIN	0x01

typedef struct {

	uint_fast16_t regid;
	/* The registration ID. */

	uint_fast16_t len;
	/* The length of the data in bytes. */

	unsigned char *data;
	/* The data. */

} jpc_com_t;

/**************************************\
* SOP marker segment parameters.
\**************************************/

typedef struct {
	uint_fast16_t seqno;
} jpc_sop_t;

/**************************************\
* Marker segment parameters.
\**************************************/

typedef struct {

	uint_fast16_t hoff;
	uint_fast16_t voff;

} jpc_crgcomp_t;

typedef struct {

	int numcomps;
	jpc_crgcomp_t *comps;

} jpc_crg_t;

typedef struct {

	unsigned char *data;
	uint_fast16_t len;

} jpc_unk_t;

typedef union {
	int soc;	/* unused */
	jpc_sot_t sot;
	int sod;	/* unused */
	int eoc;	/* unused */
	jpc_siz_t siz;
	jpc_cod_t cod;
	jpc_coc_t coc;
	jpc_rgn_t rgn;
	jpc_qcd_t qcd;
	jpc_qcc_t qcc;
	jpc_poc_t poc;
	/* jpc_plm_t plm; */
	/* jpc_plt_t plt; */
	jpc_ppm_t ppm;
	jpc_ppt_t ppt;
	jpc_sop_t sop;
	int eph;	/* unused */
	jpc_com_t com;
	jpc_crg_t crg;
	jpc_unk_t unk;
} jpc_msparms_t;

/**************************************\
* Marker segment.
\**************************************/

/* Marker segment IDs. */

#define	JPC_MS_MIN	0xff00
#define	JPC_MS_MAX	0xffff

#define	JPC_MS_INMIN	0xff80
/* The minimum marker value that cannot occur within packet data. */
#define	JPC_MS_INMAX	0xffff
/* The maximum marker value that cannot occur within packet data. */

/* Delimiting marker segments. */
#define	JPC_MS_SOC	0xff4f /* Start of code stream. */
#define	JPC_MS_SOT	0xff90 /* Start of tile-part. */
#define	JPC_MS_SOD	0xff93 /* Start of data. */
#define	JPC_MS_EOC	0xffd9 /* End of code stream. */

/* Fixed information marker segments. */
#define	JPC_MS_SIZ	0xff51 /* Image and tile size. */

/* Functional marker segments. */
#define	JPC_MS_COD	0xff52 /* Coding style default. */
#define JPC_MS_COC	0xff53 /* Coding style component. */
#define	JPC_MS_RGN	0xff5e /* Region of interest. */
#define JPC_MS_QCD	0xff5c /* Quantization default. */
#define JPC_MS_QCC	0xff5d /* Quantization component. */
#define JPC_MS_POC	0xff5f /* Progression order default. */

/* Pointer marker segments. */
#define	JPC_MS_TLM	0xff55 /* Tile-part lengths, main header. */
#define	JPC_MS_PLM	0xff57 /* Packet length, main header. */
#define	JPC_MS_PLT	0xff58 /* Packet length, tile-part header. */
#define	JPC_MS_PPM	0xff60 /* Packed packet headers, main header. */
#define	JPC_MS_PPT	0xff61 /* Packet packet headers, tile-part header. */

/* In bit stream marker segments. */
#define	JPC_MS_SOP	0xff91	/* Start of packet. */
#define	JPC_MS_EPH	0xff92	/* End of packet header. */

/* Informational marker segments. */
#define	JPC_MS_CRG	0xff63 /* Component registration. */
#define JPC_MS_COM	0xff64 /* Comment. */

struct jpc_msops_s;

typedef struct {

	uint_fast16_t id;
	/* The type of marker segment. */

	uint_fast16_t len;
	/* The length of the marker segment. */

	uint_fast32_t off;

	jpc_msparms_t parms;
	/* The parameters of the marker segment. */

	struct jpc_msops_s *ops;

} jpc_ms_t;

typedef struct jpc_msops_s {
	void (*destroyparms)(jpc_ms_t *ms);
	int (*getparms)(jpc_ms_t *ms, jpc_cstate_t *cstate, stream_t *in);
	int (*putparms)(jpc_ms_t *ms, jpc_cstate_t *cstate, stream_t *out);
	int (*dumpparms)(jpc_ms_t *ms, FILE *out);
} jpc_msops_t;

/******************************************************************************\
* Functions.
\******************************************************************************/

jpc_cstate_t *jpc_cstate_create();
void jpc_cstate_destroy(jpc_cstate_t *cstate);
jpc_ms_t *jpc_getms(stream_t *in, jpc_cstate_t *cstate);
int jpc_putms(stream_t *out, jpc_cstate_t *cstate, jpc_ms_t *ms);
int jpc_putdata(stream_t *out, stream_t *in, long n);
int jpc_getdata(stream_t *in, stream_t *out, long n);
jpc_ms_t *jpc_ms_create(int type);
void jpc_ms_destroy(jpc_ms_t *ms);

#define	JPC_MS_HASPARMS(x) \
	(!((x) == JPC_MS_SOC || (x) == JPC_MS_SOD || (x) == JPC_MS_EOC || \
	  (x) == JPC_MS_EPH || ((x) >= 0xff30 && (x) <= 0xff3f)))

#define	jpc_ms_gettype(ms) \
	((ms)->id)

int jpc_setdgblvl(int dbglvl);

int jpc_putuint16(stream_t *out, uint_fast16_t val);
int jpc_putuint32(stream_t *out, uint_fast32_t val);
int jpc_putuint8(stream_t *out, uint_fast8_t val);
int jpc_getuint32(stream_t *in, uint_fast32_t *val);
int jpc_getuint16(stream_t *in, uint_fast16_t *val);
int jpc_getuint8(stream_t *in, uint_fast8_t *val);

/* Dump a marker segment (for debugging). */
void jpc_ms_dump(jpc_ms_t *ms, FILE *out);

int jpc_setdbglevel(int dbglevel);

#endif
