/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "types.h"
#include "jpcdec.h"
#include "cs.h"
#include "mct.h"
#include "memalloc.h"
#include "t2dec.h"
#include "t1dec.h"
#include "util.h"
#include "tvparser.h"

/******************************************************************************\
*
\******************************************************************************/

#define	DIVFLOOR(x, y)	(assert(x >= 0 && y > 0), (x) / (y))
#define	DIVCEIL(x, y)	(assert(x >= 0 && y > 0), ((x) + (y) - 1) / (y))

#define	DIVPOW2FLOOR(x, y) \
	(assert(x >= 0 && y > 0), (x) >> (y))
#define	DIVPOW2CEIL(x, y) \
	(assert(x >= 0 && y >= 0), ((x) + (1 << (y)) - 1) >> (y))

/******************************************************************************\
*
\******************************************************************************/

#define	JPD_MHSOC	0x0001
  /* In the main header, expecting a SOC marker segment. */
#define	JPD_MHSIZ	0x0002
  /* In the main header, expecting a SIZ marker segment. */
#define	JPD_MH		0x0004
  /* In the main header, expecting "other" marker segments. */
#define	JPD_TPHSOT	0x0008
  /* In a tile-part header, expecting a SOT marker segment. */
#define	JPD_TPH		0x0010
  /* In a tile-part header, expecting "other" marker segments. */
#define	JPD_MT		0x0020
  /* In the main trailer. */

typedef struct {

	uint_fast16_t id;
	/* The marker segment type. */

	int validstates;
	/* The states in which this type of marker segment can be
	  validly encountered. */

	int (*action)(jpd_t *dec, jpc_ms_t *ms);
	/* The action to take upon encountering this type of marker segment. */

} jpd_mstabent_t;

/******************************************************************************\
*
\******************************************************************************/

#define	CSET	0x0001
#define	QSET	0x0002
#define	COC	0x0004
#define	QCC	0x0008

/******************************************************************************\
* Local function prototypes.
\******************************************************************************/

void dec_dump(jpd_t *dec);

ppxstab_t *ppxstab_create();
void ppxstab_destroy(ppxstab_t *tab);
int ppxstab_grow(ppxstab_t *tab, int maxents);
int ppxstab_insert(ppxstab_t *tab, ppxstabent_t *ent);
streamlist_t *ppmstabtostreams(ppxstab_t *tab);
int pptstabwrite(stream_t *out, ppxstab_t *tab);
ppxstabent_t *ppxstabent_create();
void ppxstabent_destroy(ppxstabent_t *ent);

int streamlist_numstreams(streamlist_t *streamlist);
streamlist_t *streamlist_create();
int streamlist_insert(streamlist_t *streamlist, int streamno,
  stream_t *stream);
stream_t *streamlist_remove(streamlist_t *streamlist, int streamno);
void streamlist_destroy(streamlist_t *streamlist);
stream_t *streamlist_get(streamlist_t *streamlist, int streamno);

static void jpd_cp_resetflags(jpd_cp_t *cp);
static jpd_cp_t *jpd_cp_create(uint_fast16_t numcomps);
static int jpd_cp_isvalid(jpd_cp_t *cp);
static jpd_cp_t *jpd_cp_copy(jpd_cp_t *cp);
static int jpd_cp_setfromcod(jpd_cp_t *cp, jpc_cod_t *cod);
static int jpd_cp_setfromcoc(jpd_cp_t *cp, jpc_coc_t *coc);
static int jpd_cp_setfromcox(jpd_cp_t *cp, jpd_ccp_t *ccp,
  jpc_coxcp_t *compparms, int flags);
static int jpd_cp_setfromqcd(jpd_cp_t *cp, jpc_qcd_t *qcd);
static int jpd_cp_setfromqcc(jpd_cp_t *cp, jpc_qcc_t *qcc);
static int jpd_cp_setfromqcx(jpd_cp_t *cp, jpd_ccp_t *ccp,
  jpc_qcxcp_t *compparms, int flags);
static int jpd_cp_setfromrgn(jpd_cp_t *cp, jpc_rgn_t *rgn);
static int jpd_cp_prepare(jpd_cp_t *cp);
static void jpd_cp_destroy(jpd_cp_t *cp);
static int jpd_cp_setfrompoc(jpd_cp_t *cp, jpc_poc_t *poc, int reset);
static int pi_addpchgfrompoc(pi_t *pi, jpc_poc_t *poc);

static int jpd_decode(jpd_t *dec);
static jpd_t *dec_create(jp_importopts_t *impopts, stream_t *in);
static void dec_destroy(jpd_t *dec);
static void dequantize(matrix_t *x, fix_t absstepsize);
static void undo_roi(matrix_t *x, int roishift);
static fix_t calcabsstepsize(int stepsize, int numbits);
static int jpd_tiledecode(jpd_t *dec, jpd_tile_t *tile);
static int jpd_tileinit(jpd_t *dec, jpd_tile_t *tile);
static int jpd_tilefini(jpd_t *dec, jpd_tile_t *tile);
static int jpd_process_soc(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_sot(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_sod(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_eoc(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_siz(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_cod(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_coc(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_rgn(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_qcd(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_qcc(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_poc(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_ppm(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_ppt(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_com(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_unk(jpd_t *dec, jpc_ms_t *ms);
static int jpd_process_crg(jpd_t *dec, jpc_ms_t *ms);
static int parseopts(char *optstr, jp_importopts_t *opts);

/******************************************************************************\
* Global data.
\******************************************************************************/

jpd_mstabent_t jpd_mstab[] = {
	{JPC_MS_SOC, JPD_MHSOC, jpd_process_soc},
	{JPC_MS_SOT, JPD_MH | JPD_TPHSOT, jpd_process_sot},
	{JPC_MS_SOD, JPD_TPH, jpd_process_sod},
	{JPC_MS_EOC, JPD_TPHSOT, jpd_process_eoc},
	{JPC_MS_SIZ, JPD_MHSIZ, jpd_process_siz},
	{JPC_MS_COD, JPD_MH | JPD_TPH, jpd_process_cod},
	{JPC_MS_COC, JPD_MH | JPD_TPH, jpd_process_coc},
	{JPC_MS_RGN, JPD_MH | JPD_TPH, jpd_process_rgn},
	{JPC_MS_QCD, JPD_MH | JPD_TPH, jpd_process_qcd},
	{JPC_MS_QCC, JPD_MH | JPD_TPH, jpd_process_qcc},
	{JPC_MS_POC, JPD_MH | JPD_TPH, jpd_process_poc},
	{JPC_MS_TLM, JPD_MH, 0},
	{JPC_MS_PLM, JPD_MH, 0},
	{JPC_MS_PLT, JPD_TPH, 0},
	{JPC_MS_PPM, JPD_MH, jpd_process_ppm},
	{JPC_MS_PPT, JPD_TPH, jpd_process_ppt},
	{JPC_MS_SOP, 0, 0},
	{JPC_MS_CRG, JPD_MH, jpd_process_crg},
	{JPC_MS_COM, JPD_MH | JPD_TPH, jpd_process_com},
	{0, JPD_MH | JPD_TPH, jpd_process_unk}
};

/******************************************************************************\
* The main entry point for the JPEG-2000 decoder.
\******************************************************************************/

image_t *image_load_jpc(stream_t *in, char *optstr)
{
	jp_importopts_t opts;
	jpd_t *dec;
	image_t *image;

	dec = 0;

	if (parseopts(optstr, &opts)) {
		goto error;
	}

	initluts();

	if (!(dec = dec_create(&opts, in))) {
		goto error;
	}

	jpc_setdbglevel(opts.dbglevel);
	setdbglevel(opts.dbglevel);
	memsetdbglvl(opts.dbglevel);

	if (jpd_decode(dec)) {
		goto error;
	}

	/* Save the return value. */
	image = dec->image;

	/* Stop the image from being discarded. */
	dec->image = 0;

	/* Destroy decoder. */
	dec_destroy(dec);

	return image;

error:
	if (dec) {
		dec_destroy(dec);
	}
	return 0;
}

typedef enum {
	OPT_MAXLYRS,
	OPT_DEBUG
} optid_t;

taginfo_t decopts[] = {
	{OPT_MAXLYRS, "maxlyrs"},
	{OPT_DEBUG, "debug"},
	{-1, 0}
};

static int parseopts(char *optstr, jp_importopts_t *opts)
{
	tvparser_t *tvp;

	opts->dbglevel = 0;
	opts->maxlyrs = JPC_MAXLYRS;

	if (!(tvp = tvparser_create(optstr ? optstr : ""))) {
		return -1;
	}

	while (!tvparser_next(tvp)) {
		switch (taginfo_nonull(taginfos_lookup(decopts,
		  tvparser_gettag(tvp)))->id) {
		case OPT_MAXLYRS:
			opts->maxlyrs = atoi(tvparser_getval(tvp));
			break;
		case OPT_DEBUG:
			opts->dbglevel = atoi(tvparser_getval(tvp));
			break;
		default:
			fprintf(stderr, "warning: ignoring invalid option %s\n",
			  tvparser_gettag(tvp));
			break;
		}
	}

	tvparser_destroy(tvp);

	return 0;
}

/******************************************************************************\
* Code for table-driven code stream decoder.
\******************************************************************************/

jpd_mstabent_t *jpd_mstab_lookup(uint_fast16_t id)
{
	jpd_mstabent_t *mstabent;
	for (mstabent = jpd_mstab; mstabent->id != 0; ++mstabent) {
		if (mstabent->id == id) {
			break;
		}
	}
	return mstabent;
}

static int jpd_decode(jpd_t *dec)
{
	jpc_ms_t *ms;
	jpd_mstabent_t *mstabent;
	int ret;
	jpc_cstate_t *cstate;

	if (!(cstate = jpc_cstate_create())) {
		return -1;
	}
	dec->cstate = cstate;

	/* Initially, we should expect to encounter a SOC marker segment. */
	dec->state = JPD_MHSOC;

	for (;;) {

		/* Get the next marker segment in the code stream. */
		if (!(ms = jpc_getms(dec->in, cstate))) {
			fprintf(stderr, "cannot get marker segment\n");
			return -1;
		}

		mstabent = jpd_mstab_lookup(ms->id);
		assert(mstabent);

		/* Ensure that this type of marker segment is permitted
		  at this point in the code stream. */
		if (!(dec->state & mstabent->validstates)) {
			fprintf(stderr, "unexpected marker segment type\n");
			jpc_ms_destroy(ms);
			return -1;
		}

		/* Process the marker segment. */
		if (mstabent->action) {
			ret = (*mstabent->action)(dec, ms);
		} else {
			/* No explicit action is required. */
			ret = 0;
		}

		/* Destroy the marker segment. */
		jpc_ms_destroy(ms);

		if (ret < 0) {
			return -1;
		} else if (ret > 0) {
			break;
		}

	}

	return 0;
}

static int jpd_process_crg(jpd_t *dec, jpc_ms_t *ms)
{
	uint_fast16_t cmptno;
	dec_cmpt_t *cmpt;
	jpc_crg_t *crg;

	crg = &ms->parms.crg;
	for (cmptno = 0, cmpt = dec->cmpts; cmptno < dec->numcomps; ++cmptno,
	  ++cmpt) {
		/* Ignore the information in the CRG marker segment for now.
		  This information serves no useful purpose for decoding anyhow.
		  Some other parts of the code need to be changed if these lines
		  are uncommented.
		cmpt->hsubstep = crg->comps[cmptno].hoff;
		cmpt->vsubstep = crg->comps[cmptno].voff;
		*/
	}
	return 0;
}

static int jpd_process_soc(jpd_t *dec, jpc_ms_t *ms)
{
	/* We should expect to encounter a SIZ marker segment next. */
	dec->state = JPD_MHSIZ;

	return 0;
}

static int jpd_process_sot(jpd_t *dec, jpc_ms_t *ms)
{
	jpd_tile_t *tile;
	jpc_sot_t *sot = &ms->parms.sot;
	image_cmptparm_t *compinfos;
	image_cmptparm_t *compinfo;
	dec_cmpt_t *cmpt;
	uint_fast16_t cmptno;

if (dec->state == JPD_MH) {

	compinfos = malloc(dec->numcomps * sizeof(image_cmptparm_t));
	assert(compinfos);
	for (cmptno = 0, cmpt = dec->cmpts, compinfo = compinfos;
	  cmptno < dec->numcomps; ++cmptno, ++cmpt, ++compinfo) {
		compinfo->tlx = 0;
		compinfo->tly = 0;
		compinfo->prec = cmpt->prec;
		compinfo->sgnd = cmpt->sgnd;
		compinfo->width = cmpt->width;
		compinfo->height = cmpt->height;
		compinfo->hstep = cmpt->hstep;
		compinfo->vstep = cmpt->vstep;
	}

	if (!(dec->image = image_create(dec->numcomps, compinfos,
	  CM_UNKNOWN))) {
		return -1;
	}
	free(compinfos);

	/* Is the packet header information stored in PPM marker segments in
	  the main header? */
	if (dec->ppmstab) {
		/* Convert the PPM marker segment data into a collection of streams
		  (one stream per tile-part). */
		if (!(dec->pkthdrstreams = ppmstabtostreams(dec->ppmstab))) {
			abort();
		}
		ppxstab_destroy(dec->ppmstab);
		dec->ppmstab = 0;
	}
}

	if (sot->len > 0) {
		dec->curtileendoff = stream_getrwcount(dec->in) - ms->len -
		  4 + sot->len;
	} else {
		dec->curtileendoff = 0;
	}

	if (sot->tileno > dec->numtiles) {
		fprintf(stderr, "invalid tile number in SOT marker segment\n");
		return -1;
	}
	/* Set the current tile. */
	dec->curtile = &dec->tiles[sot->tileno];
	tile = dec->curtile;
	/* Ensure that this is the expected part number. */
	if (sot->partno != tile->partno) {
		return -1;
	}
	if (tile->numparts > 0 && sot->partno >= tile->numparts) {
		return -1;
	}
	if (!tile->numparts && sot->numparts > 0) {
		tile->numparts = sot->numparts;
	}

	tile->pptstab = 0;

	switch (tile->state) {
	case JPD_TILE_INIT:
		/* This is the first tile-part for this tile. */
		tile->state = JPD_TILE_ACTIVE;
		assert(!tile->cp);
		if (!(tile->cp = jpd_cp_copy(dec->cp))) {
			return -1;
		}
		jpd_cp_resetflags(dec->cp);
		break;
	default:
		if (sot->numparts == sot->partno - 1) {
			tile->state = JPD_TILE_ACTIVELAST;
		}
		break;
	}

	/* Note: We do not increment the expected tile-part number until
	  all processing for this tile-part is complete. */

	/* We should expect to encounter other tile-part header marker
	  segments next. */
	dec->state = JPD_TPH;

	return 0;
}

static int jpd_process_sod(jpd_t *dec, jpc_ms_t *ms)
{
	jpd_tile_t *tile;
	int pos;

	if (!(tile = dec->curtile)) {
		return -1;
	}

	if (!tile->partno) {
		if (!jpd_cp_isvalid(tile->cp)) {
			return -1;
		}
		jpd_cp_prepare(tile->cp);
		if (jpd_tileinit(dec, tile)) {
			return -1;
		}
	}

	/* Are packet headers stored in the main header or tile-part header? */
	if (dec->pkthdrstreams) {
		/* Get the stream containing the packet header data for this
		  tile-part. */
		if (!(tile->pkthdrstream = streamlist_remove(dec->pkthdrstreams, 0))) {
			return -1;
		}
	}

	if (tile->pptstab) {
		if (!tile->pkthdrstream) {
			if (!(tile->pkthdrstream = stream_memopen(0, 0))) {
				return -1;
			}
		}
		pos = stream_tell(tile->pkthdrstream);
		stream_seek(tile->pkthdrstream, 0, SEEK_END);
		if (pptstabwrite(tile->pkthdrstream, tile->pptstab)) {
			return -1;
		}
		stream_seek(tile->pkthdrstream, pos, SEEK_SET);
		ppxstab_destroy(tile->pptstab);
		tile->pptstab = 0;
	}

	if (getdbglevel() >= 10) {
		dec_dump(dec);
	}

	if (dec_decodepkts(dec, (tile->pkthdrstream) ? tile->pkthdrstream :
	  dec->in, dec->in)) {
		fprintf(stderr, "dec_decodepkts failed\n");
		return -1;
	}

	/* Gobble any unconsumed tile data. */
	if (dec->curtileendoff > 0) {
		uint_fast32_t curoff;
		uint_fast32_t n;
		curoff = stream_getrwcount(dec->in);
		if (curoff < dec->curtileendoff) {
			n = dec->curtileendoff - curoff;
			fprintf(stderr,
			  "warning: ignoring trailing garbage (%lu bytes)\n",
			  (unsigned long) n);

			while (n-- > 0) {
				if (stream_getc(dec->in) == EOF) {
					fprintf(stderr, "read error\n");
					return -1;
				}
			}
		} else if (curoff > dec->curtileendoff) {
			fprintf(stderr,
			  "warning: not enough tile data (%lu bytes)\n",
			  (unsigned long) curoff - dec->curtileendoff);
		}

	}

	if (tile->numparts > 0 && tile->partno == tile->numparts - 1) {
		if (jpd_tiledecode(dec, tile)) {
			return -1;
		}
		jpd_tilefini(dec, tile);
	}

	dec->curtile = 0;

	/* Increment the expected tile-part number. */
	++tile->partno;

	/* We should expect to encounter a SOT marker segment next. */
	dec->state = JPD_TPHSOT;

	return 0;
}

static int jpd_tileinit(jpd_t *dec, jpd_tile_t *tile)
{
	jpd_tcomp_t *tcomp;
	uint_fast16_t compno;
	int rlvlno;
	jpd_rlvl_t *rlvl;
	jpd_band_t *band;
	jpd_prc_t *prc;
	int bndno;
	tsfb_band_t *bnd;
	int bandno;
	jpd_ccp_t *ccp;
	int prccnt;
	jpd_cblk_t *cblk;
	int cblkcnt;
	uint_fast32_t tlprcxstart;
	uint_fast32_t tlprcystart;
	uint_fast32_t brprcxend;
	uint_fast32_t brprcyend;
	uint_fast32_t tlcbgxstart;
	uint_fast32_t tlcbgystart;
	uint_fast32_t brcbgxend;
	uint_fast32_t brcbgyend;
	uint_fast32_t cbgxstart;
	uint_fast32_t cbgystart;
	uint_fast32_t cbgxend;
	uint_fast32_t cbgyend;
	uint_fast32_t tlcblkxstart;
	uint_fast32_t tlcblkystart;
	uint_fast32_t brcblkxend;
	uint_fast32_t brcblkyend;
	uint_fast32_t cblkxstart;
	uint_fast32_t cblkystart;
	uint_fast32_t cblkxend;
	uint_fast32_t cblkyend;
	uint_fast32_t tmpxstart;
	uint_fast32_t tmpystart;
	uint_fast32_t tmpxend;
	uint_fast32_t tmpyend;
	jpd_cp_t *cp;
	tsfb_band_t bnds[64];
	pchg_t *pchg;
	int pchgno;
	dec_cmpt_t *cmpt;

	cp = tile->cp;
	tile->realmode = 0;
	if (cp->mctid == MCT_ICT) {
		tile->realmode = 1;
	}

	for (compno = 0, tcomp = tile->tcomps, cmpt = dec->cmpts; compno <
	  dec->numcomps; ++compno, ++tcomp, ++cmpt) {
		ccp = &tile->cp->ccps[compno];
		if (ccp->qmfbid == JPC_COX_INS) {
			tile->realmode = 1;
		}
		tcomp->numrlvls = ccp->numrlvls;
		if (!(tcomp->rlvls = malloc(tcomp->numrlvls *
		  sizeof(jpd_rlvl_t)))) {
			return -1;
		}
		if (!(tcomp->data = seq2d_create(DIVCEIL(tile->xstart,
		  cmpt->hstep), DIVCEIL(tile->ystart, cmpt->vstep),
		  DIVCEIL(tile->xend, cmpt->hstep), DIVCEIL(tile->yend,
		  cmpt->vstep)))) {
			return -1;
		}
		if (!(tcomp->tsfb = cod_gettsfb(ccp->qmfbid,
		  tcomp->numrlvls - 1))) {
			return -1;
		}
{
	tsfb_getbands(tcomp->tsfb, seq2d_xstart(tcomp->data), seq2d_ystart(tcomp->data), seq2d_xend(tcomp->data), seq2d_yend(tcomp->data), bnds);
}
		for (rlvlno = 0, rlvl = tcomp->rlvls; rlvlno < tcomp->numrlvls;
		  ++rlvlno, ++rlvl) {
rlvl->bands = 0;
			rlvl->xstart = DIVPOW2CEIL(tcomp->xstart,
			  tcomp->numrlvls - 1 - rlvlno);
			rlvl->ystart = DIVPOW2CEIL(tcomp->ystart,
			  tcomp->numrlvls - 1 - rlvlno);
			rlvl->xend = DIVPOW2CEIL(tcomp->xend,
			  tcomp->numrlvls - 1 - rlvlno);
			rlvl->yend = DIVPOW2CEIL(tcomp->yend,
			  tcomp->numrlvls - 1 - rlvlno);
			rlvl->prcwidthexpn = ccp->prcwidthexpns[rlvlno];
			rlvl->prcheightexpn = ccp->prcheightexpns[rlvlno];
			tlprcxstart = DIVPOW2FLOOR(rlvl->xstart,
			  rlvl->prcwidthexpn) << rlvl->prcwidthexpn;
			tlprcystart = DIVPOW2FLOOR(rlvl->ystart,
			  rlvl->prcheightexpn) << rlvl->prcheightexpn;
			brprcxend = DIVPOW2CEIL(rlvl->xend,
			  rlvl->prcwidthexpn) << rlvl->prcwidthexpn;
			brprcyend = DIVPOW2CEIL(rlvl->yend,
			  rlvl->prcheightexpn) << rlvl->prcheightexpn;
			rlvl->numhprcs = (brprcxend - tlprcxstart) >>
			  rlvl->prcwidthexpn;
			rlvl->numvprcs = (brprcyend - tlprcystart) >>
			  rlvl->prcheightexpn;
			rlvl->numprcs = rlvl->numhprcs * rlvl->numvprcs;

			if (rlvl->xstart >= rlvl->xend || rlvl->ystart >= rlvl->yend) {
				rlvl->bands = 0;
				rlvl->numprcs = 0;
				rlvl->numhprcs = 0;
				rlvl->numvprcs = 0;
				continue;
			}	
			if (!rlvlno) {
				tlcbgxstart = tlprcxstart;
				tlcbgystart = tlprcystart;
				brcbgxend = brprcxend;
				brcbgyend = brprcyend;
				rlvl->cbgwidthexpn = rlvl->prcwidthexpn;
				rlvl->cbgheightexpn = rlvl->prcheightexpn;
			} else {
				tlcbgxstart = DIVPOW2CEIL(tlprcxstart, 1);
				tlcbgystart = DIVPOW2CEIL(tlprcystart, 1);
				brcbgxend = DIVPOW2CEIL(brprcxend, 1);
				brcbgyend = DIVPOW2CEIL(brprcyend, 1);
				rlvl->cbgwidthexpn = rlvl->prcwidthexpn - 1;
				rlvl->cbgheightexpn = rlvl->prcheightexpn - 1;
			}
			rlvl->cblkwidthexpn = min(ccp->cblkwidthexpn,
			  rlvl->cbgwidthexpn);
			rlvl->cblkheightexpn = min(ccp->cblkheightexpn,
			  rlvl->cbgheightexpn);

			rlvl->numbands = (!rlvlno) ? 1 : 3;
			if (!(rlvl->bands = malloc(rlvl->numbands *
			  sizeof(jpd_band_t)))) {
				return -1;
			}
			for (bandno = 0, band = rlvl->bands;
			  bandno < rlvl->numbands; ++bandno, ++band) {
				bndno = (!rlvlno) ? 0 : (3 * (rlvlno - 1) +
				  bandno + 1);
				bnd = &bnds[bndno];

				band->orient = bnd->orient;
				band->stepsize = ccp->stepsizes[bndno];
				band->analgain = NOMINALGAIN(ccp->qmfbid,
				  tcomp->numrlvls - 1, rlvlno, band->orient);
				band->absstepsize = calcabsstepsize(band->stepsize,
				  cmpt->prec + band->analgain);
				band->numbps = ccp->numguardbits +
				  JPC_QCX_GETEXPN(band->stepsize) - 1;
				band->roishift = (ccp->roishift + band->numbps > PREC - 1) ?
				  (PREC - 1 - band->numbps) : ccp->roishift;
				band->data = 0;
				band->prcs = 0;
				if (bnd->xstart == bnd->xend || bnd->ystart == bnd->yend) {
					continue;
				}
				if (!(band->data = seq2d_create(0, 0, 0, 0))) {
					return -1;
				}
				seq2d_bindsub(band->data, tcomp->data, bnd->locxstart, bnd->locystart, bnd->locxend, bnd->locyend);
				seq2d_setshift(band->data, bnd->xstart, bnd->ystart);

				assert(rlvl->numprcs);

				if (!(band->prcs = malloc(rlvl->numprcs * sizeof(jpd_prc_t)))) {
					return -1;
				}

/************************************************/
	cbgxstart = tlcbgxstart;
	cbgystart = tlcbgystart;
	for (prccnt = rlvl->numprcs, prc = band->prcs;
	  prccnt > 0; --prccnt, ++prc) {
		cbgxend = cbgxstart + (1 << rlvl->cbgwidthexpn);
		cbgyend = cbgystart + (1 << rlvl->cbgheightexpn);
		prc->xstart = max(cbgxstart, seq2d_xstart(band->data));
		prc->ystart = max(cbgystart, seq2d_ystart(band->data));
		prc->xend = min(cbgxend, seq2d_xend(band->data));
		prc->yend = min(cbgyend, seq2d_yend(band->data));
		if (prc->xend > prc->xstart && prc->yend > prc->ystart) {
			tlcblkxstart = DIVPOW2FLOOR(prc->xstart,
			  rlvl->cblkwidthexpn) << rlvl->cblkwidthexpn;
			tlcblkystart = DIVPOW2FLOOR(prc->ystart,
			  rlvl->cblkheightexpn) << rlvl->cblkheightexpn;
			brcblkxend = DIVPOW2CEIL(prc->xend,
			  rlvl->cblkwidthexpn) << rlvl->cblkwidthexpn;
			brcblkyend = DIVPOW2CEIL(prc->yend,
			  rlvl->cblkheightexpn) << rlvl->cblkheightexpn;
			prc->numhcblks = (brcblkxend - tlcblkxstart) >>
			  rlvl->cblkwidthexpn;
			prc->numvcblks = (brcblkyend - tlcblkystart) >>
			  rlvl->cblkheightexpn;
			prc->numcblks = prc->numhcblks * prc->numvcblks;
			assert(prc->numcblks > 0);

			if (!(prc->incltagtree = tagtree_create(prc->numhcblks, prc->numvcblks))) {
				return -1;
			}
			if (!(prc->numimsbstagtree = tagtree_create(prc->numhcblks, prc->numvcblks))) {
				return -1;
			}
			if (!(prc->cblks = malloc(prc->numcblks * sizeof(jpd_cblk_t)))) {
				return -1;
			}

			cblkxstart = cbgxstart;
			cblkystart = cbgystart;
			for (cblkcnt = prc->numcblks, cblk = prc->cblks; cblkcnt > 0;) {
				cblkxend = cblkxstart + (1 << rlvl->cblkwidthexpn);
				cblkyend = cblkystart + (1 << rlvl->cblkheightexpn);
				tmpxstart = max(cblkxstart, prc->xstart);
				tmpystart = max(cblkystart, prc->ystart);
				tmpxend = min(cblkxend, prc->xend);
				tmpyend = min(cblkyend, prc->yend);
				if (tmpxend > tmpxstart && tmpyend > tmpystart) {
					cblk->firstpassno = -1;
					cblk->mqdec = 0;
					cblk->nulldec = 0;
					cblk->flags = 0;
					cblk->numpasses = 0;
					cblk->segs.head = 0;
					cblk->segs.tail = 0;
					cblk->curseg = 0;
					cblk->numimsbs = 0;
					cblk->numlenbits = 3;
					cblk->flags = 0;
					if (!(cblk->data = seq2d_create(0, 0, 0, 0))) {
						return -1;
					}
					seq2d_bindsub(cblk->data, band->data, tmpxstart, tmpystart, tmpxend, tmpyend);
					++cblk;
					--cblkcnt;
				}
				cblkxstart += 1 << rlvl->cblkwidthexpn;
				if (cblkxstart >= cbgxend) {
					cblkxstart = cbgxstart;
					cblkystart += 1 << rlvl->cblkheightexpn;
				}
			}

		} else {
			prc->cblks = 0;
			prc->incltagtree = 0;
			prc->numimsbstagtree = 0;
		}
		cbgxstart += 1 << rlvl->cbgwidthexpn;
		if (cbgxstart >= brcbgxend) {
			cbgxstart = tlcbgxstart;
			cbgystart += 1 << rlvl->cbgheightexpn;
		}

	}
/********************************************/
			}
		}
	}

if (!(tile->pi = pi_create(dec, tile)))
{
	return -1;
}

	for (pchgno = 0; pchgno < pchglist_numpchgs(tile->cp->pchglist);
	  ++pchgno) {
		pchg = pchg_copy(pchglist_get(tile->cp->pchglist, pchgno));
		assert(pchg);
		pi_addpchg(tile->pi, pchg);
	}
	pi_init(tile->pi);

	return 0;
}

static int jpd_tilefini(jpd_t *dec, jpd_tile_t *tile)
{
	jpd_tcomp_t *tcomp;
	int compno;
	int bandno;
	int rlvlno;
	jpd_band_t *band;
	jpd_rlvl_t *rlvl;
	int prcno;
	jpd_prc_t *prc;
	jpd_seg_t *seg;
	jpd_cblk_t *cblk;
	int cblkno;

	if (tile->state == JPD_TILE_INIT || tile->state == JPD_TILE_DONE) {
		return 0;
	}

	for (compno = 0, tcomp = tile->tcomps; compno < dec->numcomps;
	  ++compno, ++tcomp) {
		for (rlvlno = 0, rlvl = tcomp->rlvls; rlvlno < tcomp->numrlvls;
		  ++rlvlno, ++rlvl) {
if (!rlvl->bands) {
	continue;
}
			for (bandno = 0, band = rlvl->bands; bandno < rlvl->numbands; ++bandno, ++band) {
if (band->prcs) {
				for (prcno = 0, prc = band->prcs; prcno <
				  rlvl->numprcs; ++prcno, ++prc) {
if (!prc->cblks) {
	continue;
}
					for (cblkno = 0, cblk = prc->cblks; cblkno < prc->numcblks; ++cblkno, ++cblk) {

	while (cblk->segs.head) {
		seg = cblk->segs.head;
		seglist_remove(&cblk->segs, seg);
		seg_destroy(seg);
	}
	matrix_destroy(cblk->data);
	if (cblk->mqdec) {
		mqdec_destroy(cblk->mqdec);
	}
	if (cblk->nulldec) {
		bitstream_close(cblk->nulldec);
	}
	if (cblk->flags) {
		matrix_destroy(cblk->flags);
	}
					}
					if (prc->incltagtree) {
						tagtree_destroy(prc->incltagtree);
					}
					if (prc->numimsbstagtree) {
						tagtree_destroy(prc->numimsbstagtree);
					}
					if (prc->cblks) {
						free(prc->cblks);
					}
				}
}
				if (band->data) {
					matrix_destroy(band->data);
				}
				if (band->prcs) {
					free(band->prcs);
				}
			}
			if (rlvl->bands) {
				free(rlvl->bands);
			}
		}
		if (tcomp->rlvls) {
			free(tcomp->rlvls);
		}
		if (tcomp->data) {
			matrix_destroy(tcomp->data);
		}
		if (tcomp->tsfb) {
			tsfb_destroy(tcomp->tsfb);
		}
	}
	if (tile->cp) {
		jpd_cp_destroy(tile->cp);
	}
	if (tile->tcomps) {
		free(tile->tcomps);
	}
	if (tile->pi) {
		pi_destroy(tile->pi);
	}

	tile->state = JPD_TILE_DONE;

	return 0;
}

static int jpd_tiledecode(jpd_t *dec, jpd_tile_t *tile)
{
	int i;
	int j;
	jpd_tcomp_t *tcomp;
	jpd_rlvl_t *rlvl;
	jpd_band_t *band;
	int compno;
	int rlvlno;
	int bandno;
	int adjust;
	int v;
	jpd_ccp_t *ccp;
	dec_cmpt_t *cmpt;

	if (dec_decodecblks(dec, tile)) {
		fprintf(stderr, "dec_decodeallsegs failed\n");
		return -1;
	}

	/* Perform dequantization. */
	for (compno = 0, tcomp = tile->tcomps; compno < dec->numcomps;
	  ++compno, ++tcomp) {
		ccp = &tile->cp->ccps[compno];
		for (rlvlno = 0, rlvl = tcomp->rlvls; rlvlno < tcomp->numrlvls;
		  ++rlvlno, ++rlvl) {
			if (!rlvl->bands) {
				continue;
			}
			for (bandno = 0, band = rlvl->bands;
			  bandno < rlvl->numbands; ++bandno, ++band) {
if (!band->data) {
	continue;
}
				undo_roi(band->data, band->roishift);
				if (tile->realmode) {
					matrix_asl(band->data, FIX_FRACBITS);
					dequantize(band->data, band->absstepsize);
				}
			}
		}
	}

	/* Apply an inverse wavelet transform if necessary. */
	for (compno = 0, tcomp = tile->tcomps; compno < dec->numcomps;
	  ++compno, ++tcomp) {
		ccp = &tile->cp->ccps[compno];
		tsfb_synthesize(tcomp->tsfb, ((ccp->qmfbid ==
		  JPC_COX_RFT) ? TSFB_RITIMODE : 0), tcomp->data);
	}


	/* Apply an inverse intercomponent transform if necessary. */
	switch (tile->cp->mctid) {
	case MCT_RCT:
		assert(dec->numcomps == 3);
		irct(tile->tcomps[0].data, tile->tcomps[1].data,
		  tile->tcomps[2].data);
		break;
	case MCT_ICT:
		assert(dec->numcomps == 3);
		iict(tile->tcomps[0].data, tile->tcomps[1].data,
		  tile->tcomps[2].data);
		break;
	}

	/* Perform rounding and convert to integer values. */
	if (tile->realmode) {
		for (compno = 0, tcomp = tile->tcomps; compno < dec->numcomps;
		  ++compno, ++tcomp) {
			for (i = 0; i < matrix_numrows(tcomp->data); ++i) {
				for (j = 0; j < matrix_numcols(tcomp->data); ++j) {
					v = matrix_get(tcomp->data, i, j);
					v = fix_round(v);
					matrix_set(tcomp->data, i, j, fixtoint(v));
				}
			}
		}
	}

	/* Perform level shift. */
	for (compno = 0, tcomp = tile->tcomps, cmpt = dec->cmpts; compno <
	  dec->numcomps; ++compno, ++tcomp, ++cmpt) {
		adjust = cmpt->sgnd ? 0 : (1 << (cmpt->prec - 1));
		for (i = 0; i < matrix_numrows(tcomp->data); ++i) {
			for (j = 0; j < matrix_numcols(tcomp->data); ++j) {
				*matrix_getref(tcomp->data, i, j) += adjust;
			}
		}
	}

	/* Perform clipping. */
	for (compno = 0, tcomp = tile->tcomps, cmpt = dec->cmpts; compno <
	  dec->numcomps; ++compno, ++tcomp, ++cmpt) {
		fix_t mn;
		fix_t mx;
		mn = cmpt->sgnd ? (-(1 << (cmpt->prec - 1))) : (0);
		mx = cmpt->sgnd ? ((1 << (cmpt->prec - 1)) - 1) : ((1 <<
		  cmpt->prec) - 1);
		matrix_clip(tcomp->data, mn, mx);
	}

	/* XXX need to free tsfb struct */

	/* Write the data for each component of the image. */
	for (compno = 0, tcomp = tile->tcomps, cmpt = dec->cmpts; compno <
	  dec->numcomps; ++compno, ++tcomp, ++cmpt) {
		if (image_writecmpt(dec->image, compno, tcomp->xstart -
		  DIVCEIL(dec->xstart, cmpt->hstep), tcomp->ystart -
		  DIVCEIL(dec->ystart, cmpt->vstep), matrix_numcols(
		  tcomp->data), matrix_numrows(tcomp->data), tcomp->data)) {
			fprintf(stderr, "write component failed\n");
			return -4;
		}
	}

	return 0;
}

static int jpd_process_eoc(jpd_t *dec, jpc_ms_t *ms)
{
	int tileno;
	jpd_tile_t *tile;
	for (tileno = 0, tile = dec->tiles; tileno < dec->numtiles; ++tileno,
	  ++tile) {
		if (tile->state == JPD_TILE_ACTIVE) {
			if (jpd_tiledecode(dec, tile)) {
				return -1;
			}
		}
		jpd_tilefini(dec, tile);
	}

	/* We are done processing the code stream. */
	dec->state = JPD_MT;

	return 1;
}

static int jpd_process_siz(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_siz_t *siz = &ms->parms.siz;
	uint_fast16_t compno;
	uint_fast32_t tileno;
	jpd_tile_t *tile;
	jpd_tcomp_t *tcomp;
	uint_fast32_t htileno;
	uint_fast32_t vtileno;
	dec_cmpt_t *cmpt;

	dec->xstart = siz->xoff;
	dec->ystart = siz->yoff;
	dec->xend = siz->width;
	dec->yend = siz->height;
	dec->tilewidth = siz->tilewidth;
	dec->tileheight = siz->tileheight;
	dec->tilexoff = siz->tilexoff;
	dec->tileyoff = siz->tileyoff;
	dec->numcomps = siz->numcomps;
	if (!(dec->cp = jpd_cp_create(dec->numcomps))) {
		return -1;
	}

	if (!(dec->cmpts = malloc(dec->numcomps * sizeof(dec_cmpt_t)))) {
		return -1;
	}

	for (compno = 0, cmpt = dec->cmpts; compno < dec->numcomps; ++compno,
	  ++cmpt) {
		cmpt->prec = siz->comps[compno].prec;
		cmpt->sgnd = siz->comps[compno].sgnd;
		cmpt->hstep = siz->comps[compno].hsamp;
		cmpt->vstep = siz->comps[compno].vsamp;
		cmpt->width = DIVCEIL(dec->xend, cmpt->hstep) -
		  DIVCEIL(dec->xstart, cmpt->hstep);
		cmpt->height = DIVCEIL(dec->yend, cmpt->vstep) -
		  DIVCEIL(dec->ystart, cmpt->vstep);
		cmpt->hsubstep = 0;
		cmpt->vsubstep = 0;
	}

	dec->image = 0;

	dec->numhtiles = DIVCEIL(dec->xend - dec->tilexoff, dec->tilewidth);
	dec->numvtiles = DIVCEIL(dec->yend - dec->tileyoff, dec->tileheight);
	dec->numtiles = dec->numhtiles * dec->numvtiles;
	if (!(dec->tiles = malloc(dec->numtiles * sizeof(jpd_tile_t)))) {
		return -1;
	}

	for (tileno = 0, tile = dec->tiles; tileno < dec->numtiles; ++tileno,
	  ++tile) {
		htileno = tileno % dec->numhtiles;
		vtileno = tileno / dec->numhtiles;
		tile->realmode = 0;
		tile->state = JPD_TILE_INIT;
		tile->xstart = max(dec->tilexoff + htileno * dec->tilewidth,
		  dec->xstart);
		tile->ystart = max(dec->tileyoff + vtileno * dec->tileheight,
		  dec->ystart);
		tile->xend = min(dec->tilexoff + (htileno + 1) *
		  dec->tilewidth, dec->xend);
		tile->yend = min(dec->tileyoff + (vtileno + 1) *
		  dec->tileheight, dec->yend);
		tile->numparts = 0;
		tile->partno = 0;
		tile->pkthdrstream = 0;
		tile->pkthdrstreampos = 0;
		tile->pptstab = 0;
		tile->cp = 0;
		if (!(tile->tcomps = malloc(dec->numcomps *
		  sizeof(jpd_tcomp_t)))) {
			return -1;
		}
		for (compno = 0, cmpt = dec->cmpts, tcomp = tile->tcomps;
		  compno < dec->numcomps; ++compno, ++cmpt, ++tcomp) {
			tcomp->rlvls = 0;
			tcomp->data = 0;
			tcomp->xstart = DIVCEIL(tile->xstart, cmpt->hstep);
			tcomp->ystart = DIVCEIL(tile->ystart, cmpt->vstep);
			tcomp->xend = DIVCEIL(tile->xend, cmpt->hstep);
			tcomp->yend = DIVCEIL(tile->yend, cmpt->vstep);
			tcomp->tsfb = 0;
		}
	}

	dec->pkthdrstreams = 0;

	/* We should expect to encounter other main header marker segments
	  or an SOT marker segment next. */
	dec->state = JPD_MH;

	return 0;
}

static int jpd_process_cod(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_cod_t *cod = &ms->parms.cod;
	jpd_tile_t *tile;

	switch (dec->state) {
	case JPD_MH:
		jpd_cp_setfromcod(dec->cp, cod);
		break;
	case JPD_TPH:
		if (!(tile = dec->curtile)) {
			return -1;
		}
		if (tile->partno != 0) {
			return -1;
		}
		jpd_cp_setfromcod(tile->cp, cod);
		break;
	}
	return 0;
}

static int jpd_process_coc(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_coc_t *coc = &ms->parms.coc;
	jpd_tile_t *tile;

	if (coc->compno > dec->numcomps) {
		fprintf(stderr,
		  "invalid component number in COC marker segment\n");
		return -1;
	}
	switch (dec->state) {
	case JPD_MH:
		jpd_cp_setfromcoc(dec->cp, coc);
		break;
	case JPD_TPH:
		if (!(tile = dec->curtile)) {
			return -1;
		}
		if (tile->partno > 0) {
			return -1;
		}
		jpd_cp_setfromcoc(tile->cp, coc);
		break;
	}
	return 0;
}

static int jpd_process_rgn(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_rgn_t *rgn = &ms->parms.rgn;
	jpd_tile_t *tile;

	if (rgn->compno > dec->numcomps) {
		fprintf(stderr,
		  "invalid component number in RGN marker segment\n");
		return -1;
	}
	switch (dec->state) {
	case JPD_MH:
		jpd_cp_setfromrgn(dec->cp, rgn);
		break;
	case JPD_TPH:
		if (!(tile = dec->curtile)) {
			return -1;
		}
		if (tile->partno > 0) {
			return -1;
		}
		jpd_cp_setfromrgn(tile->cp, rgn);
		break;
	}

	return 0;
}

static int jpd_process_qcd(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_qcd_t *qcd = &ms->parms.qcd;
	jpd_tile_t *tile;

	switch (dec->state) {
	case JPD_MH:
		jpd_cp_setfromqcd(dec->cp, qcd);
		break;
	case JPD_TPH:
		if (!(tile = dec->curtile)) {
			return -1;
		}
		if (tile->partno > 0) {
			return -1;
		}
		jpd_cp_setfromqcd(tile->cp, qcd);
		break;
	}
	return 0;
}

static int jpd_process_qcc(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_qcc_t *qcc = &ms->parms.qcc;
	jpd_tile_t *tile;

	if (qcc->compno > dec->numcomps) {
		fprintf(stderr,
		  "invalid component number in QCC marker segment\n");
		return -1;
	}
	switch (dec->state) {
	case JPD_MH:
		jpd_cp_setfromqcc(dec->cp, qcc);
		break;
	case JPD_TPH:
		if (!(tile = dec->curtile)) {
			return -1;
		}
		if (tile->partno > 0) {
			return -1;
		}
		jpd_cp_setfromqcc(tile->cp, qcc);
		break;
	}
	return 0;
}

static int jpd_process_poc(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_poc_t *poc = &ms->parms.poc;
	jpd_tile_t *tile;
	switch (dec->state) {
	case JPD_MH:
		if (jpd_cp_setfrompoc(dec->cp, poc, 1)) {
			return -1;
		}
		break;
	case JPD_TPH:
		if (!(tile = dec->curtile)) {
			return -1;
		}
		if (!tile->partno) {
			if (jpd_cp_setfrompoc(tile->cp, poc, (!tile->partno))) {
				return -1;
			}
		} else {
			pi_addpchgfrompoc(tile->pi, poc);
		}
		break;
	}
	return 0;
}

static int jpd_process_ppm(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_ppm_t *ppm = &ms->parms.ppm;
	ppxstabent_t *ppmstabent;

	if (!dec->ppmstab) {
		if (!(dec->ppmstab = ppxstab_create())) {
			return -1;
		}
	}

	if (!(ppmstabent = ppxstabent_create())) {
		return -1;
	}
	ppmstabent->ind = ppm->ind;
	ppmstabent->data = ppm->data;
	ppm->data = 0;
	ppmstabent->len = ppm->len;
	if (ppxstab_insert(dec->ppmstab, ppmstabent)) {
		return -1;
	}
	return 0;
}

static int jpd_process_ppt(jpd_t *dec, jpc_ms_t *ms)
{
	jpc_ppt_t *ppt = &ms->parms.ppt;
	jpd_tile_t *tile;
	ppxstabent_t *pptstabent;

	tile = dec->curtile;
	if (!tile->pptstab) {
		if (!(tile->pptstab = ppxstab_create())) {
			return -1;
		}
	}
	if (!(pptstabent = ppxstabent_create())) {
		return -1;
	}
	pptstabent->ind = ppt->ind;
	pptstabent->data = ppt->data;
	ppt->data = 0;
	pptstabent->len = ppt->len;
	if (ppxstab_insert(tile->pptstab, pptstabent)) {
		return -1;
	}
	return 0;
}

static int jpd_process_com(jpd_t *dec, jpc_ms_t *ms)
{
	return 0;
}

static int jpd_process_unk(jpd_t *dec, jpc_ms_t *ms)
{
	fprintf(stderr, "warning: ignoring unknown marker segment\n");
	jpc_ms_dump(ms, stderr);
	return 0;
}

/******************************************************************************\
*
\******************************************************************************/

static jpd_cp_t *jpd_cp_create(uint_fast16_t numcomps)
{
	jpd_cp_t *cp;
	jpd_ccp_t *ccp;
	int compno;

	if (!(cp = malloc(sizeof(jpd_cp_t)))) {
		return 0;
	}
	cp->flags = 0;
	cp->numcomps = numcomps;
	cp->prgord = 0;
	cp->numlyrs = 0;
	cp->mctid = 0;
	cp->csty = 0;
	if (!(cp->ccps = malloc(cp->numcomps * sizeof(jpd_ccp_t)))) {
		return 0;
	}
	if (!(cp->pchglist = pchglist_create())) {
		free(cp->ccps);
		return 0;
	}
	for (compno = 0, ccp = cp->ccps; compno < cp->numcomps;
	  ++compno, ++ccp) {
		ccp->flags = 0;
		ccp->numrlvls = 0;
		ccp->cblkwidthexpn = 0;
		ccp->cblkheightexpn = 0;
		ccp->qmfbid = 0;
		ccp->numstepsizes = 0;
		ccp->numguardbits = 0;
		ccp->roishift = 0;
		ccp->cblkctx = 0;
	}
	return cp;
}

static jpd_cp_t *jpd_cp_copy(jpd_cp_t *cp)
{
	jpd_cp_t *newcp;
	jpd_ccp_t *newccp;
	jpd_ccp_t *ccp;
	int compno;

	if (!(newcp = jpd_cp_create(cp->numcomps))) {
		return 0;
	}
	newcp->flags = cp->flags;
	newcp->prgord = cp->prgord;
	newcp->numlyrs = cp->numlyrs;
	newcp->mctid = cp->mctid;
	newcp->csty = cp->csty;
	pchglist_destroy(newcp->pchglist);
	newcp->pchglist = 0;
	if (!(newcp->pchglist = pchglist_copy(cp->pchglist))) {
		free(newcp);
		return 0;
	}
	for (compno = 0, newccp = newcp->ccps, ccp = cp->ccps;
	  compno < cp->numcomps;
	  ++compno, ++newccp, ++ccp) {
		*newccp = *ccp;
	}
	return newcp;
}

static void jpd_cp_resetflags(jpd_cp_t *cp)
{
	int compno;
	jpd_ccp_t *ccp;
	cp->flags &= (CSET | QSET);
	for (compno = 0, ccp = cp->ccps; compno < cp->numcomps;
	  ++compno, ++ccp) {
		ccp->flags = 0;
	}
}

static void jpd_cp_destroy(jpd_cp_t *cp)
{
	if (cp->ccps) {
		free(cp->ccps);
	}
	if (cp->pchglist) {
		pchglist_destroy(cp->pchglist);
	}
	free(cp);
}

static int jpd_cp_isvalid(jpd_cp_t *cp)
{
	uint_fast16_t compcnt;
	jpd_ccp_t *ccp;

	if (!(cp->flags & CSET) || !(cp->flags & QSET)) {
		return 0;
	}
	for (compcnt = cp->numcomps, ccp = cp->ccps; compcnt > 0; --compcnt,
	  ++ccp) {
		/* Is there enough step sizes for the number of bands? */
		if ((ccp->qsty != JPC_QCX_SIQNT && ccp->numstepsizes < 3 *
		  ccp->numrlvls - 2) || (ccp->qsty == JPC_QCX_SIQNT &&
		  ccp->numstepsizes != 1)) {
			return 0;
		}
	}
	return 1;
}

static void calcstepsizes(uint_fast16_t refstepsize, int numrlvls,
  uint_fast16_t *stepsizes)
{
	int bandno;
	int numbands;
	uint_fast16_t expn;
	uint_fast16_t mant;
	expn = JPC_QCX_GETEXPN(refstepsize);
	mant = JPC_QCX_GETMANT(refstepsize);
	numbands = 3 * numrlvls - 2;
	for (bandno = 0; bandno < numbands; ++bandno) {
		stepsizes[bandno] = JPC_QCX_MANT(mant) | JPC_QCX_EXPN(expn +
		  (numrlvls - 1) - (numrlvls - 1 - ((bandno > 0) ? ((bandno + 2) / 3) : (0))));
	}
}

static int jpd_cp_prepare(jpd_cp_t *cp)
{
	jpd_ccp_t *ccp;
	int compno;
	int i;
	for (compno = 0, ccp = cp->ccps; compno < cp->numcomps;
	  ++compno, ++ccp) {
		if (!(ccp->csty & JPC_COX_PRT)) {
			for (i = 0; i < JPC_MAXRLVLS; ++i) {
				ccp->prcwidthexpns[i] = 15;
				ccp->prcheightexpns[i] = 15;
			}
		}
		if (ccp->qsty == JPC_QCX_SIQNT) {
			calcstepsizes(ccp->stepsizes[0], ccp->numrlvls, ccp->stepsizes);
		}
	}
	return 0;
}

static int jpd_cp_setfromcod(jpd_cp_t *cp, jpc_cod_t *cod)
{
	jpd_ccp_t *ccp;
	int compno;
	cp->flags |= CSET;
	cp->prgord = cod->prg;
	if (cod->mctrans) {
		cp->mctid = (cod->compparms.qmfbid == JPC_COX_INS) ? (MCT_ICT) : (MCT_RCT);
	} else {
		cp->mctid = MCT_NONE;
	}
	cp->numlyrs = cod->numlyrs;
	cp->csty = cod->csty & (JPC_COD_SOP | JPC_COD_EPH);
	for (compno = 0, ccp = cp->ccps; compno < cp->numcomps;
	  ++compno, ++ccp) {
		jpd_cp_setfromcox(cp, ccp, &cod->compparms, 0);
	}
	cp->flags |= CSET;
	return 0;
}

static int jpd_cp_setfromcoc(jpd_cp_t *cp, jpc_coc_t *coc)
{
	jpd_cp_setfromcox(cp, &cp->ccps[coc->compno], &coc->compparms, COC);
	return 0;
}

static int jpd_cp_setfromcox(jpd_cp_t *cp, jpd_ccp_t *ccp,
  jpc_coxcp_t *compparms, int flags)
{
	int rlvlno;
	if ((flags & COC) || !(ccp->flags & COC)) {
		ccp->numrlvls = compparms->numdlvls + 1;
		ccp->cblkwidthexpn = JPC_COX_GETCBLKSIZEEXPN(
		  compparms->cblkwidthval);
		ccp->cblkheightexpn = JPC_COX_GETCBLKSIZEEXPN(
		  compparms->cblkheightval);
		ccp->qmfbid = compparms->qmfbid;
		ccp->cblkctx = compparms->cblksty;
		ccp->csty = compparms->csty & JPC_COX_PRT;
		for (rlvlno = 0; rlvlno < compparms->numrlvls; ++rlvlno) {
			ccp->prcwidthexpns[rlvlno] =
			  compparms->rlvls[rlvlno].parwidthval;
			ccp->prcheightexpns[rlvlno] =
			  compparms->rlvls[rlvlno].parheightval;
		}
		ccp->flags |= flags | CSET;
	}
	return 0;
}

static int jpd_cp_setfromqcd(jpd_cp_t *cp, jpc_qcd_t *qcd)
{
	int compno;
	jpd_ccp_t *ccp;
	for (compno = 0, ccp = cp->ccps; compno < cp->numcomps;
	  ++compno, ++ccp) {
		jpd_cp_setfromqcx(cp, ccp, &qcd->compparms, 0);
	}
	cp->flags |= QSET;
	return 0;
}

static int jpd_cp_setfromqcc(jpd_cp_t *cp, jpc_qcc_t *qcc)
{
	return jpd_cp_setfromqcx(cp, &cp->ccps[qcc->compno], &qcc->compparms, QCC);
}

static int jpd_cp_setfromqcx(jpd_cp_t *cp, jpd_ccp_t *ccp,
  jpc_qcxcp_t *compparms, int flags)
{
	int bandno;
	if ((flags & QCC) || !(ccp->flags & QCC)) {
		ccp->flags |= flags | QSET;
		for (bandno = 0; bandno < compparms->numstepsizes; ++bandno) {
			ccp->stepsizes[bandno] = compparms->stepsizes[bandno];
		}
		ccp->numstepsizes = compparms->numstepsizes;
		ccp->numguardbits = compparms->numguard;
		ccp->qsty = compparms->qntsty;
	}
	return 0;
}

static int jpd_cp_setfromrgn(jpd_cp_t *cp, jpc_rgn_t *rgn)
{
	jpd_ccp_t *ccp;
	ccp = &cp->ccps[rgn->compno];
	ccp->roishift = rgn->roishift;
	return 0;
}

static int pi_addpchgfrompoc(pi_t *pi, jpc_poc_t *poc)
{
	int pchgno;
	pchg_t *pchg;
	for (pchgno = 0; pchgno < poc->numpchgs; ++pchgno) {
		if (!(pchg = pchg_copy(&poc->pchgs[pchgno]))) {
			return -1;
		}
		if (pchglist_insert(pi->pchglist, -1, pchg)) {
			return -1;
		}
	}
	return 0;
}

static int jpd_cp_setfrompoc(jpd_cp_t *cp, jpc_poc_t *poc, int reset)
{
	int pchgno;
	pchg_t *pchg;
	if (reset) {
		while (pchglist_numpchgs(cp->pchglist) > 0) {
			pchg = pchglist_remove(cp->pchglist, 0);
			pchg_destroy(pchg);
		}
	}
	for (pchgno = 0; pchgno < poc->numpchgs; ++pchgno) {
		if (!(pchg = pchg_copy(&poc->pchgs[pchgno]))) {
			return -1;
		}
		if (pchglist_insert(cp->pchglist, -1, pchg)) {
			return -1;
		}
	}
	return 0;
}

static fix_t calcabsstepsize(int stepsize, int numbits)
{
	fix_t absstepsize;
	int n;

	absstepsize = inttofix(1);
	n = FIX_FRACBITS - 11;
	absstepsize |= (n >= 0) ? (JPC_QCX_GETMANT(stepsize) << n) :
	  (JPC_QCX_GETMANT(stepsize) >> (-n));
	n = numbits - JPC_QCX_GETEXPN(stepsize);
	absstepsize = (n >= 0) ? (absstepsize << n) : (absstepsize >> (-n));
	return absstepsize;
}

static void dequantize(matrix_t *x, fix_t absstepsize)
{
	int i;
	int j;
	int t;
	int r;

	if (absstepsize == inttofix(1)) {
		return;
	}

	r = 1 << (FIX_FRACBITS - 1);

	for (i = 0; i < matrix_numrows(x); ++i) {
		for (j = 0; j < matrix_numcols(x); ++j) {
			t = matrix_get(x, i, j);
			if (t) {
				t = fix_mul(fix_add(t, ((t < 0) ? (-r) : r)), absstepsize);
			} else {
				t = 0;
			}
			matrix_set(x, i, j, t);
		}
	}

}

static void undo_roi(matrix_t *x, int roishift)
{
	int i;
	int j;
	fix_t a;
	int thresh;
	fix_t b;

	if (roishift == 0) {
		return;
	}
	thresh = 1 << roishift;

	for (i = 0; i < matrix_numrows(x); ++i) {
		for (j = 0; j < matrix_numcols(x); ++j) {
			a = matrix_get(x, i, j);
			b = (a < 0) ? (-a) : a;
			if (b >= thresh /*|| !(b & 0xff)*/) {
				b >>= roishift;
				a = (a < 0) ? (-b) : b;
				matrix_set(x, i, j, a);
			}
		}
	}
}

static jpd_t *dec_create(jp_importopts_t *impopts, stream_t *in)
{
	jpd_t *dec;

	if (!(dec = malloc(sizeof(jpd_t)))) {
		return 0;
	}

	dec->image = 0;
	dec->xstart = 0;
	dec->ystart = 0;
	dec->xend = 0;
	dec->yend = 0;
	dec->tilewidth = 0;
	dec->tileheight = 0;
	dec->tilexoff = 0;
	dec->tileyoff = 0;
	dec->numhtiles = 0;
	dec->numvtiles = 0;
	dec->numtiles = 0;
	dec->tiles = 0;
	dec->curtile = 0;
	dec->numcomps = 0;
	dec->in = in;
	dec->cp = 0;
	dec->maxlyrs = impopts->maxlyrs;
	dec->ppmseqno = 0;
	dec->state = 0;
	dec->cmpts = 0;
	dec->pkthdrstreams = 0;
	dec->ppmstab = 0;
	dec->curtileendoff = 0;

	return dec;
}

static void dec_destroy(jpd_t *dec)
{
	if (dec->cstate) {
		jpc_cstate_destroy(dec->cstate);
	}
	if (dec->pkthdrstreams) {
		streamlist_destroy(dec->pkthdrstreams);
	}
	if (dec->image) {
		image_destroy(dec->image);
	}

	if (dec->cp) {
		jpd_cp_destroy(dec->cp);
	}

	if (dec->cmpts) {
		free(dec->cmpts);
	}

	if (dec->tiles) {
		free(dec->tiles);
	}

	free(dec);
}

/******************************************************************************\
*
\******************************************************************************/

void seglist_insert(jpd_seglist_t *list, jpd_seg_t *ins, jpd_seg_t *node)
{
	jpd_seg_t *prev;
	jpd_seg_t *next;

	prev = ins;
	node->prev = prev;
	next = prev ? (prev->next) : 0;
	node->prev = prev;
	node->next = next;
	if (prev) {
		prev->next = node;
	} else {
		list->head = node;
	}
	if (next) {
		next->prev = node;
	} else {
		list->tail = node;
	}
}

void seglist_remove(jpd_seglist_t *list, jpd_seg_t *seg)
{
	jpd_seg_t *prev;
	jpd_seg_t *next;

	prev = seg->prev;
	next = seg->next;
	if (prev) {
		prev->next = next;
	} else {
		list->head = next;
	}
	if (next) {
		next->prev = prev;
	} else {
		list->tail = prev;
	}
	seg->prev = 0;
	seg->next = 0;
}

jpd_seg_t *seg_alloc()
{
	jpd_seg_t *seg;

	if (!(seg = malloc(sizeof(jpd_seg_t)))) {
		return 0;
	}
	seg->prev = 0;
	seg->next = 0;
	seg->passno = -1;
	seg->numpasses = 0;
	seg->maxpasses = 0;
	seg->type = SEG_INVALID;
	seg->stream = 0;
	seg->cnt = 0;
	seg->complete = 0;
	seg->lyrno = -1;
	return seg;
}

void seg_destroy(jpd_seg_t *seg)
{
	if (seg->stream) {
		stream_close(seg->stream);
	}
	free(seg);
}

static int jpd_dump(jpd_t *dec, FILE *out)
{
	jpd_tile_t *tile;
	int tileno;
	jpd_tcomp_t *tcomp;
	uint_fast16_t compno;
	jpd_rlvl_t *rlvl;
	int rlvlno;
	jpd_band_t *band;
	int bandno;
	jpd_prc_t *prc;
	int prcno;
	jpd_cblk_t *cblk;
	int cblkno;

	for (tileno = 0, tile = dec->tiles; tileno < dec->numtiles;
	  ++tileno, ++tile) {
		for (compno = 0, tcomp = tile->tcomps; compno < dec->numcomps;
		  ++compno, ++tcomp) {
			for (rlvlno = 0, rlvl = tcomp->rlvls; rlvlno <
			  tcomp->numrlvls; ++rlvlno, ++rlvl) {
fprintf(out, "RESOLUTION LEVEL %d\n", rlvlno);
fprintf(out, "xs =%d, ys = %d, xe = %d, ye = %d, w = %d, h = %d\n",
  rlvl->xstart, rlvl->ystart, rlvl->xend, rlvl->yend, rlvl->xend -
  rlvl->xstart, rlvl->yend - rlvl->ystart);
				for (bandno = 0, band = rlvl->bands;
				  bandno < rlvl->numbands; ++bandno, ++band) {
fprintf(out, "BAND %d\n", bandno);
fprintf(out, "xs =%d, ys = %d, xe = %d, ye = %d, w = %d, h = %d\n",
  seq2d_xstart(band->data), seq2d_ystart(band->data), seq2d_xend(band->data),
  seq2d_yend(band->data), seq2d_xend(band->data) - seq2d_xstart(band->data),
  seq2d_yend(band->data) - seq2d_ystart(band->data));
					for (prcno = 0, prc = band->prcs;
					  prcno < rlvl->numprcs; ++prcno,
					  ++prc) {
fprintf(out, "CODE BLOCK GROUP %d\n", prcno);
fprintf(out, "xs =%d, ys = %d, xe = %d, ye = %d, w = %d, h = %d\n",
  prc->xstart, prc->ystart, prc->xend, prc->yend, prc->xend -
  prc->xstart, prc->yend - prc->ystart);
						for (cblkno = 0, cblk =
						  prc->cblks; cblkno <
						  prc->numcblks; ++cblkno,
						  ++cblk) {
fprintf(out, "CODE BLOCK %d\n", cblkno);
fprintf(out, "xs =%d, ys = %d, xe = %d, ye = %d, w = %d, h = %d\n",
  seq2d_xstart(cblk->data), seq2d_ystart(cblk->data), seq2d_xend(cblk->data),
  seq2d_yend(cblk->data), seq2d_xend(cblk->data) - seq2d_xstart(cblk->data),
  seq2d_yend(cblk->data) - seq2d_ystart(cblk->data));
						}
					}
				}
			}
		}
	}

	return 0;
}

streamlist_t *streamlist_create()
{
	streamlist_t *streamlist;
	int i;

	if (!(streamlist = malloc(sizeof(streamlist_t)))) {
		return 0;
	}
	streamlist->numstreams = 0;
	streamlist->maxstreams = 100;
	if (!(streamlist->streams = malloc(streamlist->maxstreams *
	  sizeof(stream_t *)))) {
		free(streamlist);
		return 0;
	}
	for (i = 0; i < streamlist->maxstreams; ++i) {
		streamlist->streams[i] = 0;
	}
	return streamlist;
}

int streamlist_insert(streamlist_t *streamlist, int streamno,
  stream_t *stream)
{
	stream_t **newstreams;
	int newmaxstreams;
	int i;
	/* Grow the array of streams if necessary. */
	if (streamlist->numstreams >= streamlist->maxstreams) {
		newmaxstreams = streamlist->maxstreams + 1024;
		if (!(newstreams = realloc(streamlist->streams,
		  (newmaxstreams + 1024) * sizeof(stream_t *)))) {
			return -1;
		}
		for (i = streamlist->numstreams; i < streamlist->maxstreams; ++i) {
			streamlist->streams[i] = 0;
		}
		streamlist->maxstreams = newmaxstreams;
		streamlist->streams = newstreams;
	}
	if (streamno != streamlist->numstreams) {
		/* Can only handle insertion at start of list. */
		return -1;
	}
	streamlist->streams[streamno] = stream;
	++streamlist->numstreams;
	return 0;
}

stream_t *streamlist_remove(streamlist_t *streamlist, int streamno)
{
	stream_t *stream;
	int i;
	if (streamno >= streamlist->numstreams) {
		abort();
	}
	stream = streamlist->streams[streamno];
	for (i = streamno + 1; i < streamlist->numstreams; ++i) {
		streamlist->streams[i - 1] = streamlist->streams[i];
	}
	--streamlist->numstreams;
	return stream;
}

void streamlist_destroy(streamlist_t *streamlist)
{
	int streamno;
	if (streamlist->streams) {
		for (streamno = 0; streamno < streamlist->numstreams;
		  ++streamno) {
			stream_close(streamlist->streams[streamno]);
		}
		free(streamlist->streams);
	}
	free(streamlist);
}

stream_t *streamlist_get(streamlist_t *streamlist, int streamno)
{
	assert(streamno < streamlist->numstreams);
	return streamlist->streams[streamno];
}

int streamlist_numstreams(streamlist_t *streamlist)
{
	return streamlist->numstreams;
}

void dec_dump(jpd_t *dec)
{
	jpd_tile_t *tile;
	jpd_tcomp_t *tcmpt;
	jpd_rlvl_t *rlvl;
	jpd_band_t *band;
	jpd_prc_t *prc;
	jpd_cblk_t *cblk;
	uint_fast16_t cmptno;
	uint_fast16_t rlvlno;
	uint_fast16_t bandno;
	uint_fast32_t prcno;
	uint_fast32_t cblkno;

	tile = &dec->tiles[0];
tile = dec->curtile;

	for (cmptno = 0, tcmpt = tile->tcomps; cmptno < dec->numcomps; ++cmptno,
	  ++tcmpt) {
		fprintf(stderr, "  tcmpt %5d %5d %5d %5d\n", seq2d_xstart(tcmpt->data), seq2d_ystart(tcmpt->data), seq2d_xend(tcmpt->data), seq2d_yend(tcmpt->data));
		for (rlvlno = 0, rlvl = tcmpt->rlvls; rlvlno < tcmpt->numrlvls;
		  ++rlvlno, ++rlvl) {
			fprintf(stderr, "    rlvl %5d %5d %5d %5d\n", rlvl->xstart, rlvl->ystart, rlvl->xend, rlvl->yend);
			for (bandno = 0, band = rlvl->bands; bandno < rlvl->numbands;
			  ++bandno, ++band) {
				if (!band->data) {
					continue;
				}
				fprintf(stderr, "      band %5d %5d %5d %5d\n", seq2d_xstart(band->data), seq2d_ystart(band->data), seq2d_xend(band->data), seq2d_yend(band->data));
				for (prcno = 0, prc = band->prcs; prcno < rlvl->numprcs;
				  ++prcno, ++prc) {
					fprintf(stderr, "        prc %5d %5d %5d %5d\n", prc->xstart, prc->ystart, prc->xend, prc->yend);
					if (!prc->cblks) {
						continue;
					}
					for (cblkno = 0, cblk = prc->cblks; cblkno < prc->numcblks;
					  ++cblkno, ++cblk) {
						fprintf(stderr, "         cblk %5d %5d %5d %5d\n", seq2d_xstart(cblk->data), seq2d_ystart(cblk->data), seq2d_xend(cblk->data), seq2d_yend(cblk->data));
					}
				}
			}
		}
	}
}

ppxstab_t *ppxstab_create()
{
	ppxstab_t *tab;

	if (!(tab = malloc(sizeof(ppxstab_t)))) {
		return 0;
	}
	tab->numents = 0;
	tab->maxents = 0;
	tab->ents = 0;
	return tab;
}

void ppxstab_destroy(ppxstab_t *tab)
{
	int i;
	for (i = 0; i < tab->numents; ++i) {
		ppxstabent_destroy(tab->ents[i]);
	}
	if (tab->ents) {
		free(tab->ents);
	}
	free(tab);
}

int ppxstab_grow(ppxstab_t *tab, int maxents)
{
	ppxstabent_t **newents;
	if (tab->maxents < maxents) {
		newents = (tab->ents) ? realloc(tab->ents, maxents *
		  sizeof(ppxstabent_t *)) : malloc(maxents * sizeof(ppxstabent_t *));
		if (!newents) {
			return -1;
		}
		tab->ents = newents;
		tab->maxents = maxents;
	}
	return 0;
}

int ppxstab_insert(ppxstab_t *tab, ppxstabent_t *ent)
{
	int inspt;
	int i;

	for (i = 0; i < tab->numents; ++i) {
		if (tab->ents[i]->ind > ent->ind) {
			break;
		}
	}
	inspt = i;

	if (tab->numents >= tab->maxents) {
		if (ppxstab_grow(tab, tab->maxents + 128)) {
			return -1;
		}
	}

	for (i = tab->numents; i > inspt; --i) {
		tab->ents[i] = tab->ents[i - 1];
	}
	tab->ents[i] = ent;
	++tab->numents;

	return 0;
}

streamlist_t *ppmstabtostreams(ppxstab_t *tab)
{
	streamlist_t *streams;
	uchar *dataptr;
	uint_fast32_t datacnt;
	uint_fast32_t tpcnt;
	ppxstabent_t *ent;
	int entno;
	stream_t *stream;
	int n;

	if (!(streams = streamlist_create())) {
		goto error;
	}

	if (!tab->numents) {
		return streams;
	}

	entno = 0;
	ent = tab->ents[entno];
	dataptr = ent->data;
	datacnt = ent->len;
	for (;;) {

		/* Get the length of the packet header data for the current
		  tile-part. */
		if (datacnt < 4) {
			goto error;
		}
		if (!(stream = stream_memopen(0, 0))) {
			goto error;
		}
		if (streamlist_insert(streams, streamlist_numstreams(streams),
		  stream)) {
			goto error;
		}
		tpcnt = (dataptr[0] << 24) | (dataptr[1] << 16) | (dataptr[2] << 8)
		  | dataptr[3];
		datacnt -= 4;
		dataptr += 4;

		/* Get the packet header data for the current tile-part. */
		while (tpcnt) {
			if (!datacnt) {
				if (++entno >= tab->numents) {
					goto error;
				}
				ent = tab->ents[entno];
				dataptr = ent->data;
				datacnt = ent->len;
			}
			n = min(tpcnt, datacnt);
			if (stream_write(stream, dataptr, n) != n) {
				goto error;
			}
			tpcnt -= n;
			dataptr += n;
			datacnt -= n;
		}
		stream_rewind(stream);
		if (!datacnt) {
			if (++entno >= tab->numents) {
				break;
			}
			ent = tab->ents[entno];
			dataptr = ent->data;
			datacnt = ent->len;
		}
	}

	return streams;

error:
	streamlist_destroy(streams);
	return 0;
}

int pptstabwrite(stream_t *out, ppxstab_t *tab)
{
	int i;
	ppxstabent_t *ent;
	for (i = 0; i < tab->numents; ++i) {
		ent = tab->ents[i];
		if (stream_write(out, ent->data, ent->len) != ent->len) {
			return -1;
		}
	}
	return 0;
}

ppxstabent_t *ppxstabent_create()
{
	ppxstabent_t *ent;
	if (!(ent = malloc(sizeof(ppxstabent_t)))) {
		return 0;
	}
	ent->data = 0;
	ent->len = 0;
	ent->ind = 0;
	return ent;
}

void ppxstabent_destroy(ppxstabent_t *ent)
{
	if (ent->data) {
		free(ent->data);
	}
	free(ent);
}
