/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * MQ Arithmetic Decoder
 *
 * $Id$
 */

#ifndef MQDEC_H
#define MQDEC_H

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "types.h"
#include "stream.h"
#include "mqcod.h"

/******************************************************************************\
* Types.
\******************************************************************************/

/* MQ arithmetic decoder. */

typedef struct {

	uint_fast32_t creg;
	/* The C register. */

	uint_fast32_t areg;
	/* The A register. */

	uint_fast32_t ctreg;
	/* The CT register. */

	mqstate_t **curctx;
	/* The current context. */

	mqstate_t **ctxs;
	/* The per-context information. */

	int maxctxs;
	/* The maximum number of contexts. */

	stream_t *in;
	/* The stream from which to read data. */

	unsigned char inbuffer;
	/* The last character read. */

	int eof;
	/* The EOF indicator. */

} mqdec_t;

/******************************************************************************\
* Functions and macros.
\******************************************************************************/

mqdec_t *mqdec_create(int maxctxs, stream_t *in);
/* Create a MQ decoder. */

void mqdec_destroy(mqdec_t *dec);
/* Destroy a MQ decoder. */

void mqdec_setinput(mqdec_t *dec, stream_t *in);
/* Set the input stream associated with a MQ decoder. */

void mqdec_init(mqdec_t *dec);
/* Initialize a MQ decoder. */

void mqdec_setctx(mqdec_t *dec, int ctxno, mqctx_t *ctx);
/* Set the state information for a particular context of a MQ decoder. */

void mqdec_setctxs(mqdec_t *dec, int numctxs, mqctx_t *ctxs);
/* Set the state information for all contexts of a MQ decoder. */

void mqdec_dump(mqdec_t *dec, FILE *out);
/* Dump the MQ decoder state for debugging. */

#define	mqdec_setcurctx(dec, ctxno) \
	((mqdec)->curctx = &(mqdec)->ctxs[ctxno]);
/* Set the current context for a MQ decoder. */

#if !defined(DEBUG)
#define	mqdec_getbit(dec) \
	mqdec_getbit_macro(dec)
#else
#define	mqdec_getbit(dec) \
	mqdec_getbit_func(dec)
#endif
/* Decode a symbol. */

#if !defined(DEBUG)
#define	mqdec_getbitnoskew(dec) \
	mqdec_getbit_macro(dec)
#else
#define	mqdec_getbitnoskew(dec) \
	mqdec_getbit_func(dec)
#endif
/* Decode a symbol (assuming an unskewed probability distribution). */

/******************************************************************************\
* EVERYTHING BELOW THIS POINT IS IMPLEMENTATION SPECIFIC AND NOT PART OF THE
* APPLICATION INTERFACE.  DO NOT RELY ON ANY OF THE INTERNAL FUNCTIONS/MACROS
* GIVEN BELOW.
\******************************************************************************/

#define	mqdec_getbit_macro(dec) \
	((((dec)->areg -= (*(dec)->curctx)->qeval), \
	  (dec)->creg >> 16 >= (*(dec)->curctx)->qeval) ? \
	  ((((dec)->creg -= (*(dec)->curctx)->qeval << 16), \
	  (dec)->areg & 0x8000) ?  (*(dec)->curctx)->mps : \
	  mqdec_mpsexchrenormd(dec)) : \
	  mqdec_lpsexchrenormd(dec))

#define	mqdec_mpsexchange(areg, delta, curctx, bit) \
{ \
	if ((areg) < (delta)) { \
		register mqstate_t *state = *(curctx); \
		/* LPS decoded. */ \
		(bit) = state->mps ^ 1; \
		*(curctx) = state->nlps; \
	} else { \
		register mqstate_t *state = *(curctx); \
		/* MPS decoded. */ \
		(bit) = state->mps; \
		*(curctx) = state->nmps; \
	} \
}

#define	mqdec_lpsexchange(areg, delta, curctx, bit) \
{ \
	if ((areg) >= (delta)) { \
		register mqstate_t *state = *(curctx); \
		(areg) = (delta); \
		(bit) = state->mps ^ 1; \
		*(curctx) = state->nlps; \
	} else { \
		register mqstate_t *state = *(curctx); \
		(areg) = (delta); \
		(bit) = state->mps; \
		*(curctx) = state->nmps; \
	} \
}

#define	mqdec_renormd(areg, creg, ctreg, in, eof, inbuf) \
{ \
	do { \
		if (!(ctreg)) { \
			mqdec_bytein2(creg, ctreg, in, eof, inbuf); \
		} \
		(areg) <<= 1; \
		(creg) <<= 1; \
		--(ctreg); \
	} while (!((areg) & 0x8000)); \
}

#define	mqdec_bytein2(creg, ctreg, in, eof, inbuf) \
{ \
	int c; \
	unsigned char prevbuf; \
	if (!(eof)) { \
		if ((c = stream_getc(in)) == EOF) { \
			(eof) = 1; \
			c = 0xff; \
		} \
		prevbuf = (inbuf); \
		(inbuf) = c; \
		if (prevbuf == 0xff) { \
			if (c > 0x8f) { \
				(creg) += 0xff00; \
				(ctreg) = 8; \
			} else { \
				(creg) += c << 9; \
				(ctreg) = 7; \
			} \
		} else { \
			(creg) += c << 8; \
			(ctreg) = 8; \
		} \
	} else { \
		(creg) += 0xff00; \
		(ctreg) = 8; \
	} \
}

int mqdec_getbit_func(mqdec_t *dec);
int mqdec_mpsexchrenormd(mqdec_t *dec);
int mqdec_lpsexchrenormd(mqdec_t *dec);

#endif
