/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Quadrature Mirror-Image Filter Bank (QMFB) Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <assert.h>

#include "qmfb.h"
#include "tsfb.h"
#include "fix.h"
#include "memalloc.h"

/******************************************************************************\
*
\******************************************************************************/

#define	DIVPOW2FLOOR(x, y) \
	(assert(x >= 0 && y > 0), (x) >> (y))
#define	DIVPOW2CEIL(x, y) \
	(assert(x >= 0 && y >= 0), ((x) + (1 << (y)) - 1) >> (y))

/******************************************************************************\
*
\******************************************************************************/

static qmfb1d_t *qmfb1d_create();

static int ft_getnumchans(qmfb1d_t *qmfb);
static int ft_getanalfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters);
static int ft_getsynfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters);
static void ft_analyze(qmfb1d_t *qmfb, int flags, seq2d_t *x);
static void ft_synthesize(qmfb1d_t *qmfb, int flags, seq2d_t *x);

static int ns_getnumchans(qmfb1d_t *qmfb);
static int ns_getanalfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters);
static int ns_getsynfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters);
static void ns_analyze(qmfb1d_t *qmfb, int flags, seq2d_t *x);
static void ns_synthesize(qmfb1d_t *qmfb, int flags, seq2d_t *x);

/******************************************************************************\
*
\******************************************************************************/

qmfb1dops_t ft_ops = {
	ft_getnumchans,
	ft_getanalfilters,
	ft_getsynfilters,
	ft_analyze,
	ft_synthesize
};

qmfb1dops_t ns_ops = {
	ns_getnumchans,
	ns_getanalfilters,
	ns_getsynfilters,
	ns_analyze,
	ns_synthesize
};

/******************************************************************************\
*
\******************************************************************************/

static void qmfb1d_setup(fix_t *startptr, int startind, int endind,
  int intrastep, fix_t **lstartptr, int *lstartind, int *lendind,
  fix_t **hstartptr, int *hstartind, int *hendind)
{
	*lstartind = DIVPOW2CEIL(startind, 1);
	*lendind = DIVPOW2CEIL(endind, 1);
	*hstartind = DIVPOW2FLOOR(startind, 1);
	*hendind = DIVPOW2FLOOR(endind, 1);
	*lstartptr = startptr;
	*hstartptr = &startptr[(*lendind - *lstartind) * intrastep];
}

static void qmfb1d_split(fix_t *startptr, int startind, int endind,
  register int step, fix_t *lstartptr, int lstartind, int lendind,
  fix_t *hstartptr, int hstartind, int hendind)
{
	int bufsize = DIVPOW2CEIL(endind - startind, 2);
#if defined(WIN32)
#define QMFB_SPLITBUFSIZE 4096
	fix_t splitbuf[QMFB_SPLITBUFSIZE];
#else
	fix_t splitbuf[bufsize];
#endif
	fix_t *buf = splitbuf;
	int llen;
	int hlen;
	int twostep;
	fix_t *tmpptr;
	register fix_t *ptr;
	register fix_t *hptr;
	register fix_t *lptr;
	register int n;
	int state;

	twostep = step << 1;
	llen = lendind - lstartind;
	hlen = hendind - hstartind;

#if defined(WIN32)
	/* Get a buffer. */
	if (bufsize > QMFB_SPLITBUFSIZE) {
		if (!(buf = malloc(bufsize * sizeof(fix_t)))) {
			/* We have no choice but to commit suicide in this case. */
			abort();
		}
	}
#endif

	if (hstartind < lstartind) {
		/* The first sample in the input signal is to appear
		  in the highpass subband signal. */
		/* Copy the appropriate samples into the lowpass subband
		  signal, saving any samples destined for the highpass subband
		  signal as they are overwritten. */
		tmpptr = buf;
		ptr = &startptr[step];
		lptr = lstartptr;
		n = llen;
		state = 1;
		while (n-- > 0) {
			if (state) {
				*tmpptr = *lptr;
				++tmpptr;
			}
			*lptr = *ptr;
			ptr += twostep;
			lptr += step;
			state ^= 1;
		}
		/* Copy the appropriate samples into the highpass subband
		  signal. */
		/* Handle the nonoverwritten samples. */
		hptr = &hstartptr[(hlen - 1) * step];
		ptr = &startptr[(((llen + hlen - 1) >> 1) << 1) * step];
		n = hlen - (tmpptr - buf);
		while (n-- > 0) {
			*hptr = *ptr;
			hptr -= step;
			ptr -= twostep;
		}
		/* Handle the overwritten samples. */
		n = tmpptr - buf;
		while (n-- > 0) {
			--tmpptr;
			*hptr = *tmpptr;
			hptr -= step;
		}
	} else {
		/* The first sample in the input signal is to appear
		  in the lowpass subband signal. */
		/* Copy the appropriate samples into the lowpass subband
		  signal, saving any samples for the highpass subband
		  signal as they are overwritten. */
		state = 0;
		ptr = startptr;
		lptr = lstartptr;
		tmpptr = buf;
		n = llen;
		while (n-- > 0) {
			if (state) {
				*tmpptr = *lptr;
				++tmpptr;
			}
			*lptr = *ptr;
			ptr += twostep;
			lptr += step;
			state ^= 1;
		}
		/* Copy the appropriate samples into the highpass subband
		  signal. */
		/* Handle the nonoverwritten samples. */
		ptr = &startptr[((((llen + hlen) >> 1) << 1) - 1) * step];
		hptr = &hstartptr[(hlen - 1) * step];
		n = hlen - (tmpptr - buf);
		while (n-- > 0) {
			*hptr = *ptr;
			ptr -= twostep;
			hptr -= step;
		}
		/* Handle the overwritten samples. */
		n = tmpptr - buf;
		while (n-- > 0) {
			--tmpptr;
			*hptr = *tmpptr;
			hptr -= step;
		}
	}

#if defined(WIN32)
	/* If the split buffer was allocated on the heap, free this memory. */
	if (buf != splitbuf) {
		free(buf);
	}
#endif
}

static void qmfb1d_join(fix_t *startptr, int startind, int endind,
  register int step, fix_t *lstartptr, int lstartind, int lendind,
  fix_t *hstartptr, int hstartind, int hendind)
{
	int bufsize = DIVPOW2CEIL(endind - startind, 2);
#if defined(WIN32)
#define	QMFB_JOINBUFSIZE	4096
	fix_t joinbuf[QMFB_JOINBUFSIZE];
#else
	fix_t joinbuf[bufsize];
#endif
	fix_t *buf = joinbuf;
	int llen;
	int hlen;
	int twostep;
	fix_t *tmpptr;
	register fix_t *ptr;
	register fix_t *hptr;
	register fix_t *lptr;
	register int n;
	int state;

#if defined(WIN32)
	/* Allocate memory for the join buffer from the heap. */
	if (bufsize > QMFB_JOINBUFSIZE) {
		if (!(buf = malloc(bufsize * sizeof(fix_t)))) {
			/* We have no choice but to commit suicide. */
			abort();
		}
	}
#endif

	twostep = step << 1;
	llen = lendind - lstartind;
	hlen = hendind - hstartind;

	if (hstartind < lstartind) {
		/* The first sample in the highpass subband signal is to
		  appear first in the output signal. */
		/* Copy the appropriate samples into the first phase of the
		  output signal. */
		tmpptr = buf;
		hptr = hstartptr;
		ptr = startptr;
		n = (llen + 1) >> 1;
		while (n-- > 0) {
			*tmpptr = *ptr;
			*ptr = *hptr;
			++tmpptr;
			ptr += twostep;
			hptr += step;
		}
		n = hlen - ((llen + 1) >> 1);
		while (n-- > 0) {
			*ptr = *hptr;
			ptr += twostep;
			hptr += step;
		}
		/* Copy the appropriate samples into the second phase of
		  the output signal. */
		ptr -= (lendind > hendind) ? (step) : (step + twostep);
		state = !((llen - 1) & 1);
		lptr = &lstartptr[(llen - 1) * step];
		n = llen;
		while (n-- > 0) {
			if (state) {
				--tmpptr;
				*ptr = *tmpptr;
			} else {
				*ptr = *lptr;
			}
			lptr -= step;
			ptr -= twostep;
			state ^= 1;
		}
	} else {
		/* The first sample in the lowpass subband signal is to
		  appear first in the output signal. */
		/* Copy the appropriate samples into the first phase of the
		  output signal (corresponding to even indexed samples). */
		lptr = &lstartptr[(llen - 1) * step];
		ptr = &startptr[((llen - 1) << 1) * step];
		n = llen >> 1;
		tmpptr = buf;
		while (n-- > 0) {
			*tmpptr = *ptr;
			*ptr = *lptr;
			++tmpptr;
			ptr -= twostep;
			lptr -= step;
		}
		n = llen - (llen >> 1);
		while (n-- > 0) {
			*ptr = *lptr;
			ptr -= twostep;
			lptr -= step;
		}
		/* Copy the appropriate samples into the second phase of
		  the output signal (corresponding to odd indexed
		  samples). */
		ptr = &startptr[step];
		hptr = hstartptr;
		state = !(llen & 1);
		n = hlen;
		while (n-- > 0) {
			if (state) {
				--tmpptr;
				*ptr = *tmpptr;
			} else {
				*ptr = *hptr;
			}
			hptr += step;
			ptr += twostep;
			state ^= 1;
		}
	}

#if defined(WIN32)
	/* If the join buffer was allocated on the heap, free this memory. */
	if (buf != joinbuf) {
		free(buf);
	}
#endif
}

/******************************************************************************\
* Code for 5/3 transform.
\******************************************************************************/

static int ft_getnumchans(qmfb1d_t *qmfb)
{
	return 2;
}

static int ft_getanalfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters)
{
	abort();
	return -1;
}

static int ft_getsynfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters)
{
	seq_t *lf;
	seq_t *hf;

	lf = 0;
	hf = 0;

	if (len > 1 || (!len)) {
		if (!(lf = seq_create(-1, 2))) {
			goto error;
		}
		seq_set(lf, -1, dbltofix(0.5));
		seq_set(lf, 0, dbltofix(1.0));
		seq_set(lf, 1, dbltofix(0.5));
		if (!(hf = seq_create(-1, 4))) {
			goto error;
		}
		seq_set(hf, -1, dbltofix(-0.125));
		seq_set(hf, 0, dbltofix(-0.25));
		seq_set(hf, 1, dbltofix(0.75));
		seq_set(hf, 2, dbltofix(-0.25));
		seq_set(hf, 3, dbltofix(-0.125));
	} else if (len == 1) {
		if (!(lf = seq_create(0, 1))) {
			goto error;
		}
		seq_set(lf, 0, dbltofix(1.0));
		if (!(hf = seq_create(0, 1))) {
			goto error;
		}
		seq_set(hf, 0, dbltofix(2.0));
	} else {
		abort();
	}

	filters[0] = lf;
	filters[1] = hf;

	return 0;

error:
	if (lf) {
		seq_destroy(lf);
	}
	if (hf) {
		seq_destroy(hf);
	}
	return -1;
}

#define	NFT_LIFT0(lstartptr, lstartind, lendind, hstartptr, hstartind, hendind, step, pluseq) \
{ \
	register fix_t *lptr = (lstartptr); \
	register fix_t *hptr = (hstartptr); \
	register int n = (hendind) - (hstartind); \
	if ((hstartind) < (lstartind)) { \
		pluseq(*hptr, *lptr); \
		hptr += (step); \
		--n; \
	} \
	if ((hendind) >= (lendind)) { \
		--n; \
	} \
	while (n-- > 0) { \
		pluseq(*hptr, fix_asr(fix_add(*lptr, lptr[(step)]), 1)); \
		hptr += (step); \
		lptr += (step); \
	} \
	if ((hendind) >= (lendind)) { \
		pluseq(*hptr, *lptr); \
	} \
}

#define	NFT_LIFT1(lstartptr, lstartind, lendind, hstartptr, hstartind, hendind, step, pluseq) \
{ \
	register fix_t *lptr = (lstartptr); \
	register fix_t *hptr = (hstartptr); \
	register int n = (lendind) - (lstartind); \
	if ((hstartind) >= (lstartind)) { \
		pluseq(*lptr, *hptr); \
		lptr += (step); \
		--n; \
	} \
	if ((lendind) > (hendind)) { \
		--n; \
	} \
	while (n-- > 0) { \
		pluseq(*lptr, fix_asr(fix_add(*hptr, hptr[(step)]), 2)); \
		lptr += (step); \
		hptr += (step); \
	} \
	if ((lendind) > (hendind)) { \
		pluseq(*lptr, *hptr); \
	} \
}

#define	RFT_LIFT0(lstartptr, lstartind, lendind, hstartptr, hstartind, hendind, step, pmeqop) \
{ \
	register fix_t *lptr = (lstartptr); \
	register fix_t *hptr = (hstartptr); \
	register int n = (hendind) - (hstartind); \
	if ((hstartind) < (lstartind)) { \
		*hptr pmeqop *lptr; \
		hptr += (step); \
		--n; \
	} \
	if ((hendind) >= (lendind)) { \
		--n; \
	} \
	while (n-- > 0) { \
		*hptr pmeqop (*lptr + lptr[(step)]) >> 1; \
		hptr += (step); \
		lptr += (step); \
	} \
	if ((hendind) >= (lendind)) { \
		*hptr pmeqop *lptr; \
	} \
}

#define	RFT_LIFT1(lstartptr, lstartind, lendind, hstartptr, hstartind, hendind, step, pmeqop) \
{ \
	register fix_t *lptr = (lstartptr); \
	register fix_t *hptr = (hstartptr); \
	register int n = (lendind) - (lstartind); \
	if ((hstartind) >= (lstartind)) { \
		*lptr pmeqop ((*hptr << 1) + 2) >> 2; \
		lptr += (step); \
		--n; \
	} \
	if ((lendind) > (hendind)) { \
		--n; \
	} \
	while (n-- > 0) { \
		*lptr pmeqop ((*hptr + hptr[(step)]) + 2) >> 2; \
		lptr += (step); \
		hptr += (step); \
	} \
	if ((lendind) > (hendind)) { \
		*lptr pmeqop ((*hptr << 1) + 2) >> 2; \
	} \
}

static void ft_analyze(qmfb1d_t *qmfb, int flags, seq2d_t *x)
{
	fix_t *startptr;
	int startind;
	int endind;
	fix_t *  lstartptr;
	int   lstartind;
	int   lendind;
	fix_t *  hstartptr;
	int   hstartind;
	int   hendind;
	int interstep;
	int intrastep;
	int numseq;

	if (flags & QMFB1D_VERT) {
		interstep = 1;
		intrastep = seq2d_rowstep(x);
		numseq = seq2d_width(x);
		startind = seq2d_ystart(x);
		endind = seq2d_yend(x);
	} else {
		interstep = seq2d_rowstep(x);
		intrastep = 1;
		numseq = seq2d_height(x);
		startind = seq2d_xstart(x);
		endind = seq2d_xend(x);
	}

	assert(startind < endind);

	startptr = seq2d_getref(x, seq2d_xstart(x), seq2d_ystart(x));
	if (flags & QMFB1D_RITIMODE) {
		while (numseq-- > 0) {
			qmfb1d_setup(startptr, startind, endind, intrastep,
			  &lstartptr, &lstartind, &lendind, &hstartptr,
			  &hstartind, &hendind);
			if (endind - startind > 1) {
				qmfb1d_split(startptr, startind, endind,
				  intrastep, lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind);
				RFT_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep, -=);
				RFT_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep, +=);
			} else {
				if (lstartind == lendind) {
					*startptr <<= 1;
				}
			}
			startptr += interstep;
		}
	} else {
		while (numseq-- > 0) {
			qmfb1d_setup(startptr, startind, endind, intrastep,
			  &lstartptr, &lstartind, &lendind, &hstartptr,
			  &hstartind, &hendind);
			if (endind - startind > 1) {
				qmfb1d_split(startptr, startind, endind,
				  intrastep, lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind);
				NFT_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  fix_minuseq);
				NFT_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  fix_pluseq);
			} else {
				if (lstartind == lendind) {
					*startptr = fix_asl(*startptr, 1);
				}
			}
			startptr += interstep;
		}
	}
}

static void ft_synthesize(qmfb1d_t *qmfb, int flags, seq2d_t *x)
{
	fix_t *startptr;
	int startind;
	int endind;
	fix_t *lstartptr;
	int lstartind;
	int lendind;
	fix_t *hstartptr;
	int hstartind;
	int hendind;
	int interstep;
	int intrastep;
	int numseq;

	if (flags & QMFB1D_VERT) {
		interstep = 1;
		intrastep = seq2d_rowstep(x);
		numseq = seq2d_width(x);
		startind = seq2d_ystart(x);
		endind = seq2d_yend(x);
	} else {
		interstep = seq2d_rowstep(x);
		intrastep = 1;
		numseq = seq2d_height(x);
		startind = seq2d_xstart(x);
		endind = seq2d_xend(x);
	}

	assert(startind < endind);

	startptr = seq2d_getref(x, seq2d_xstart(x), seq2d_ystart(x));
	if (flags & QMFB1D_RITIMODE) {
		while (numseq-- > 0) {
			qmfb1d_setup(startptr, startind, endind, intrastep,
			  &lstartptr, &lstartind, &lendind, &hstartptr,
			  &hstartind, &hendind);
			if (endind - startind > 1) {
				RFT_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep, -=);
				RFT_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep, +=);
				qmfb1d_join(startptr, startind, endind,
				  intrastep, lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind);
			} else {
				if (lstartind == lendind) {
					*startptr >>= 1;
				}
			}
			startptr += interstep;
		}
	} else {
		while (numseq-- > 0) {
			qmfb1d_setup(startptr, startind, endind, intrastep,
			  &lstartptr, &lstartind, &lendind, &hstartptr,
			  &hstartind, &hendind);
			if (endind - startind > 1) {
				NFT_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  fix_minuseq);
				NFT_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  fix_pluseq);
				qmfb1d_join(startptr, startind, endind,
				  intrastep, lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind);
			} else {
				if (lstartind == lendind) {
					*startptr = fix_asr(*startptr, 1);
				}
			}
			startptr += interstep;
		}
	}
}

/******************************************************************************\
* Code for 9/7 transform.
\******************************************************************************/

static int ns_getnumchans(qmfb1d_t *qmfb)
{
	return 2;
}

static int ns_getanalfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters)
{
	abort();
	return -1;
}

static int ns_getsynfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters)
{
	seq_t *lf;
	seq_t *hf;

	lf = 0;
	hf = 0;

	if (len > 1 || (!len)) {
		if (!(lf = seq_create(-3, 4))) {
			goto error;
		}
		seq_set(lf, -3, dbltofix(-0.09127176311424948));
		seq_set(lf, -2, dbltofix(-0.05754352622849957));
		seq_set(lf, -1, dbltofix(0.5912717631142470));
		seq_set(lf, 0, dbltofix(1.115087052456994));
		seq_set(lf, 1, dbltofix(0.5912717631142470));
		seq_set(lf, 2, dbltofix(-0.05754352622849957));
		seq_set(lf, 3, dbltofix(-0.09127176311424948));
		if (!(hf = seq_create(-3, 6))) {
			goto error;
		}
		seq_set(hf, -3, dbltofix(-0.02674875741080976 * 2.0));
		seq_set(hf, -2, dbltofix(-0.01686411844287495 * 2.0));
		seq_set(hf, -1, dbltofix(0.07822326652898785 * 2.0));
		seq_set(hf, 0, dbltofix(0.2668641184428723 * 2.0));
		seq_set(hf, 1, dbltofix(-0.6029490182363579 * 2.0));
		seq_set(hf, 2, dbltofix(0.2668641184428723 * 2.0));
		seq_set(hf, 3, dbltofix(0.07822326652898785 * 2.0));
		seq_set(hf, 4, dbltofix(-0.01686411844287495 * 2.0));
		seq_set(hf, 5, dbltofix(-0.02674875741080976 * 2.0));
	} else if (len == 1) {
		if (!(lf = seq_create(0, 1))) {
			goto error;
		}
		seq_set(lf, 0, dbltofix(1.0));
		if (!(hf = seq_create(0, 1))) {
			goto error;
		}
		seq_set(hf, 0, dbltofix(2.0));
	} else {
		abort();
	}

	filters[0] = lf;
	filters[1] = hf;

	return 0;

error:
	if (lf) {
		seq_destroy(lf);
	}
	if (hf) {
		seq_destroy(hf);
	}
	return -1;
}

#define	NNS_LIFT0(lstartptr, lstartind, lendind, hstartptr, hstartind, hendind, step, alpha) \
{ \
	register fix_t *lptr = (lstartptr); \
	register fix_t *hptr = (hstartptr); \
	register int n = (hendind) - (hstartind); \
	fix_t twoalpha = fix_mulbyint(alpha, 2); \
	if ((hstartind) < (lstartind)) { \
		fix_pluseq(*hptr, fix_mul(*lptr, (twoalpha))); \
		hptr += (step); \
		--n; \
	} \
	if ((hendind) >= (lendind)) { \
		--n; \
	} \
	while (n-- > 0) { \
		fix_pluseq(*hptr, fix_mul(fix_add(*lptr, lptr[(step)]), (alpha))); \
		hptr += (step); \
		lptr += (step); \
	} \
	if ((hendind) >= (lendind)) { \
		fix_pluseq(*hptr, fix_mul(*lptr, (twoalpha))); \
	} \
}

#define	NNS_LIFT1(lstartptr, lstartind, lendind, hstartptr, hstartind, hendind, step, alpha) \
{ \
	register fix_t *lptr = (lstartptr); \
	register fix_t *hptr = (hstartptr); \
	register int n = (lendind) - (lstartind); \
	int twoalpha = fix_mulbyint(alpha, 2); \
	if ((hstartind) >= (lstartind)) { \
		fix_pluseq(*lptr, fix_mul(*hptr, (twoalpha))); \
		lptr += (step); \
		--n; \
	} \
	if ((lendind) > (hendind)) { \
		--n; \
	} \
	while (n-- > 0) { \
		fix_pluseq(*lptr, fix_mul(fix_add(*hptr, hptr[(step)]), (alpha))); \
		lptr += (step); \
		hptr += (step); \
	} \
	if ((lendind) > (hendind)) { \
		fix_pluseq(*lptr, fix_mul(*hptr, (twoalpha))); \
	} \
}

#define	NNS_SCALE(startptr, startind, endind, step, alpha) \
{ \
	register fix_t *ptr = (startptr); \
	register int n = (endind) - (startind); \
	while (n-- > 0) { \
		fix_muleq(*ptr, alpha); \
		ptr += (step); \
	} \
}

static void ns_analyze(qmfb1d_t *qmfb, int flags, seq2d_t *x)
{
	fix_t *startptr;
	int startind;
	int endind;
	fix_t *lstartptr;
	int lstartind;
	int lendind;
	fix_t *hstartptr;
	int hstartind;
	int hendind;
	int interstep;
	int intrastep;
	int numseq;

	if (flags & QMFB1D_VERT) {
		interstep = 1;
		intrastep = seq2d_rowstep(x);
		numseq = seq2d_width(x);
		startind = seq2d_ystart(x);
		endind = seq2d_yend(x);
	} else {
		interstep = seq2d_rowstep(x);
		intrastep = 1;
		numseq = seq2d_height(x);
		startind = seq2d_xstart(x);
		endind = seq2d_xend(x);
	}

	assert(startind < endind);

	startptr = seq2d_getref(x, seq2d_xstart(x), seq2d_ystart(x));
	if (!(flags & QMFB1D_RITIMODE)) {
		while (numseq-- > 0) {
			qmfb1d_setup(startptr, startind, endind, intrastep,
			  &lstartptr, &lstartind, &lendind, &hstartptr,
			  &hstartind, &hendind);
			if (endind - startind > 1) {
				qmfb1d_split(startptr, startind, endind,
				  intrastep, lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind);
				NNS_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(-1.586134342));
				NNS_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(-0.052980118));
				NNS_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(0.882911075));
				NNS_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(0.443506852));
				NNS_SCALE(lstartptr, lstartind, lendind,
				  intrastep, dbltofix(1.0/1.23017410558578));
				NNS_SCALE(hstartptr, hstartind, hendind,
				  intrastep, dbltofix(1.0/1.62578613134411));
			} else {
#if 0
				if (lstartind == lendind) {
					*startptr = fix_asl(*startptr, 1);
				}
#endif
			}
			startptr += interstep;
		}
	} else {
		/* The reversible integer-to-integer mode is not supported
		  for this transform. */
		abort();
	}
}

static void ns_synthesize(qmfb1d_t *qmfb, int flags, seq2d_t *x)
{
	fix_t *startptr;
	int startind;
	int endind;
	fix_t *lstartptr;
	int lstartind;
	int lendind;
	fix_t *hstartptr;
	int hstartind;
	int hendind;
	int interstep;
	int intrastep;
	int numseq;

	if (flags & QMFB1D_VERT) {
		interstep = 1;
		intrastep = seq2d_rowstep(x);
		numseq = seq2d_width(x);
		startind = seq2d_ystart(x);
		endind = seq2d_yend(x);
	} else {
		interstep = seq2d_rowstep(x);
		intrastep = 1;
		numseq = seq2d_height(x);
		startind = seq2d_xstart(x);
		endind = seq2d_xend(x);
	}

	assert(startind < endind);

	startptr = seq2d_getref(x, seq2d_xstart(x), seq2d_ystart(x));
	if (!(flags & QMFB1D_RITIMODE)) {
		while (numseq-- > 0) {
			qmfb1d_setup(startptr, startind, endind, intrastep,
			  &lstartptr, &lstartind, &lendind, &hstartptr,
			  &hstartind, &hendind);
			if (endind - startind > 1) {
#if !defined(DO_NOT_USE_NS)
				NNS_SCALE(lstartptr, lstartind, lendind,
				  intrastep, dbltofix(1.23017410558578));
				NNS_SCALE(hstartptr, hstartind, hendind,
				  intrastep, dbltofix(1.62578613134411));
				NNS_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(-0.443506852));
				NNS_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(-0.882911075));
				NNS_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(0.052980118));
				NNS_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  dbltofix(1.586134342));
#else
				NNS_SCALE(lstartptr, lstartind, lendind,
				  intrastep, dbltofix(1.0));
				NNS_SCALE(hstartptr, hstartind, hendind,
				  intrastep, dbltofix(-0.5));
				NFT_LIFT1(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  fix_minuseq);
				NFT_LIFT0(lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind, intrastep,
				  fix_pluseq);
#endif
				qmfb1d_join(startptr, startind, endind,
				  intrastep, lstartptr, lstartind, lendind,
				  hstartptr, hstartind, hendind);
			} else {
#if 0
				if (lstartind == lendind) {
					*startptr = fix_asr(*startptr, 1);
				}
#endif
			}
			startptr += interstep;
		}
	} else {
		/* The reversible integer-to-integer mode is not supported
		  for this transform. */
		abort();
	}
}

/******************************************************************************\
\******************************************************************************/

/******************************************************************************\
********************************************************************************
********************************************************************************
********************************************************************************
\******************************************************************************/

/******************************************************************************\
*
\******************************************************************************/



/******************************************************************************\
*
\******************************************************************************/

qmfb1d_t *qmfb1d_make(int qmfbid)
{
	qmfb1d_t *qmfb;
	if (!(qmfb = qmfb1d_create())) {
		return 0;
	}
	switch (qmfbid) {
	case QMFB1D_FT:
		qmfb->ops = &ft_ops;
		break;
	case QMFB1D_NS:
		qmfb->ops = &ns_ops;
		break;
	default:
		qmfb1d_destroy(qmfb);
		return 0;
		break;
	}
	return qmfb;
}

static qmfb1d_t *qmfb1d_create()
{
	qmfb1d_t *qmfb;
	if (!(qmfb = malloc(sizeof(qmfb1d_t)))) {
		return 0;
	}
	qmfb->ops = 0;
	return qmfb;
}

qmfb1d_t *qmfb1d_copy(qmfb1d_t *qmfb)
{
	qmfb1d_t *newqmfb;

	if (!(newqmfb = qmfb1d_create())) {
		return 0;
	}
	newqmfb->ops = qmfb->ops;
	return newqmfb;
}

void qmfb1d_destroy(qmfb1d_t *qmfb)
{
	free(qmfb);
}

/******************************************************************************\
*
\******************************************************************************/

void qmfb1d_getbands(qmfb1d_t *qmfb, int flags, uint_fast32_t xstart,
  uint_fast32_t ystart, uint_fast32_t xend, uint_fast32_t yend, int maxbands,
  int *numbandsptr, qmfb1dband_t *bands)
{
	int start;
	int end;
	if (flags & QMFB1D_VERT) {
		start = ystart;
		end = yend;
	} else {
		start = xstart;
		end = xend;
	}
	assert(qmfb1d_getnumchans(qmfb) == 2);
	assert(start <= end);
	bands[0].start = DIVPOW2CEIL(start, 1);
	bands[0].end = DIVPOW2CEIL(end, 1);
	bands[0].locstart = start;
	bands[0].locend = start + bands[0].end - bands[0].start;
	bands[1].start = DIVPOW2FLOOR(start, 1);
	bands[1].end = DIVPOW2FLOOR(end, 1);
	bands[1].locstart = bands[0].locend;
	bands[1].locend = bands[1].locstart + bands[1].end - bands[1].start;
	assert(bands[1].locend == end);
	*numbandsptr = 2;
}

/******************************************************************************\
*
\******************************************************************************/

int qmfb1d_getnumchans(qmfb1d_t *qmfb)
{
	return (*qmfb->ops->getnumchans)(qmfb);
}

int qmfb1d_getanalfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters)
{
	return (*qmfb->ops->getanalfilters)(qmfb, len, filters);
}

int qmfb1d_getsynfilters(qmfb1d_t *qmfb, int len, seq2d_t **filters)
{
	return (*qmfb->ops->getsynfilters)(qmfb, len, filters);
}

void qmfb1d_analyze(qmfb1d_t *qmfb, int flags, seq2d_t *x)
{
	(*qmfb->ops->analyze)(qmfb, flags, x);
}

void qmfb1d_synthesize(qmfb1d_t *qmfb, int flags, seq2d_t *x)
{
	(*qmfb->ops->synthesize)(qmfb, flags, x);
}
