/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Sun Rasterfile Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <assert.h>
#include <stdlib.h>

#include "ras.h"
#include "memalloc.h"

/******************************************************************************\
* Prototypes.
\******************************************************************************/

static int ras_puthdr(stream_t *out, ras_hdr_t *hdr);
static int ras_putint(stream_t *out, int val);

static int ras_putdata(stream_t *out, ras_hdr_t *hdr, image_t *image);
static int ras_putdatastd(stream_t *out, ras_hdr_t *hdr, image_t *image);

/******************************************************************************\
* Code.
\******************************************************************************/

int image_save_ras(image_t *image, stream_t *out, char *optstr)
{
	uint_fast32_t width;
	uint_fast32_t height;
	uint_fast32_t depth;
	uint_fast16_t cmptno;
	uint_fast16_t numcmpts;
	int i;
	ras_hdr_t hdr;
	int rowsize;

	if (optstr) {
		fprintf(stderr, "warning: ignoring RAS encoder options\n");
	}

	numcmpts = image_numcmpts(image);
	width = image_cmptwidth(image, 0);
	height = image_cmptheight(image, 0);
	depth = image_cmptprec(image, 0);

	for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
		if (image_cmptwidth(image, cmptno) != width ||
		  image_cmptheight(image, cmptno) != height ||
		  image_cmptprec(image, cmptno) != depth ||
		  image_cmptsgnd(image, cmptno) != false ||
		  image_cmpttlx(image, cmptno) != 0 ||
		  image_cmpttly(image, cmptno) != 0) {
			fprintf(stderr, "The RAS format cannot be used to represent an image with this geometry.\n");
			return -1;
		}
	}

	/* Ensure that the image can be encoded in the desired format. */
	if (numcmpts == 3) {

		/* All three components must have the same subsampling
		  factor and have a precision of eight bits. */
		if (numcmpts > 1) {
			for (i = 0; i < numcmpts; ++i) {
				if (image_cmptprec(image, i) != 8) {
					return -1;
				}
			}
		}
	} else if (numcmpts == 1) {
		/* NOP */
	} else {
		return -1;
	}

	hdr.magic = RAS_MAGIC;
	hdr.width = width;
	hdr.height = height;
	hdr.depth = (numcmpts == 3) ? 24 : depth;

	rowsize = RAS_ROWSIZE(&hdr);
	hdr.length = rowsize * hdr.height;
	hdr.type = RAS_TYPE_STD;

	hdr.maptype = RAS_MT_NONE;
	hdr.maplength = 0;

	if (ras_puthdr(out, &hdr)) {
		return -1;
	}

	if (ras_putdata(out, &hdr, image)) {
		return -1;
	}

	return 0;
}

static int ras_putdata(stream_t *out, ras_hdr_t *hdr, image_t *image)
{
	int ret;

	switch (hdr->type) {
	case RAS_TYPE_STD:
		ret = ras_putdatastd(out, hdr, image);
		break;
	default:
		ret = -1;
		break;
	}
	return ret;
}

static int ras_putdatastd(stream_t *out, ras_hdr_t *hdr, image_t *image)
{
	int rowsize;
	int pad;
	unsigned int z;
	int nz;
	int c;
	int x;
	int y;
	int v;
	matrix_t *data[3];
	int i;

	for (i = 0; i < image_numcmpts(image); ++i) {
		data[i] = matrix_create(image_height(image), image_width(image));
		assert(data[i]);
	}

	rowsize = RAS_ROWSIZE(hdr);
	pad = rowsize - (hdr->width * hdr->depth + 7) / 8;

	hdr->length = hdr->height * rowsize;

	for (y = 0; y < hdr->height; y++) {
		for (i = 0; i < image_numcmpts(image); ++i) {
			image_readcmpt(image, i, 0, y, image_width(image),
			  1, data[i]);
		}
		z = 0;
		nz = 0;
		for (x = 0; x < hdr->width; x++) {
			z <<= hdr->depth;
			if (RAS_ISRGB(hdr)) {
				v = RAS_RED((matrix_getv(data[0], x))) |
				  RAS_GREEN((matrix_getv(data[1], x))) |
				  RAS_BLUE((matrix_getv(data[2], x)));
			} else {
				v = (matrix_getv(data[0], x));
			}
			z |= v & RAS_ONES(hdr->depth);
			nz += hdr->depth;
			while (nz >= 8) {
				c = (z >> (nz - 8)) & 0xff;
				if (stream_putc(out, c) == EOF) {
					return -1;
				}
				nz -= 8;
				z &= RAS_ONES(nz);
			}
		}
		if (nz > 0) {
			c = (z >> (8 - nz)) & RAS_ONES(nz);
			if (stream_putc(out, c) == EOF) {
				return -1;
			}
		}
		if (pad % 2) {
			if (stream_putc(out, 0) == EOF) {
				return -1;
			}
		}
	}

	for (i = 0; i < image_numcmpts(image); ++i) {
		matrix_destroy(data[i]);
	}

	return 0;
}

static int ras_puthdr(stream_t *out, ras_hdr_t *hdr)
{
	if (ras_putint(out, RAS_MAGIC) || ras_putint(out, hdr->width) ||
	  ras_putint(out, hdr->height) || ras_putint(out, hdr->depth) ||
	  ras_putint(out, hdr->length) || ras_putint(out, hdr->type) ||
	  ras_putint(out, hdr->maptype) || ras_putint(out, hdr->maplength)) {
		return -1;
	}

	return 0;
}

static int ras_putint(stream_t *out, int val)
{
	int x;
	int i;
	int c;

	x = val;
	for (i = 0; i < 4; i++) {
		c = (x >> 24) & 0xff;
		if (stream_putc(out, c) == EOF) {
			return -1;
		}
		x <<= 8;
	}

	return 0;
}
